from __future__ import division
import logging
import os

from six import iteritems
import numpy as np
import SimpleITK as sitk
from PIL import Image, ImageDraw
from skimage.draw import polygon

from allensdk.config.manifest import Manifest


# use np_sitk_convert or sitk_np_convert to access
# TODO: check if this already exists. If not: add more dtypes
# it does not
NUMPY_SITK_TYPE_LOOKUP = {np.dtype(np.float32): sitk.sitkFloat32}
SITK_NUMPY_TYPE_LOOKUP = {v: k for k, v in iteritems(NUMPY_SITK_TYPE_LOOKUP)}


# ITK/Numpy


def set_image_spacing(image, spacing, origin=True):
    '''
    '''

    spacing = np.array(spacing)

    image.SetSpacing(spacing.tolist())

    if origin:
        image.SetOrigin( (0.5 * spacing).tolist() )


def new_image(dims, spacing, dtype, origin=True):
    '''
    '''
    
    if len(dims) == 2:
        image = sitk.Image(dims[0], dims[1], dtype)
    elif len(dims) == 3:
        image = sitk.Image(dims[0], dims[1], dims[2], dtype)
    set_image_spacing(image, spacing, origin)
    
    return image


def image_from_array(array, spacing, origin=True):
    '''
    '''

    image = sitk.GetImageFromArray(array)
    set_image_spacing(image, spacing, origin)
    
    return image


def np_sitk_convert(np_type):
    '''
    '''
    
    return NUMPY_SITK_TYPE_LOOKUP[np_type]
    
    
def sitk_np_convert(sitk_type):
    '''
    '''
    
    return SITK_NUMPY_TYPE_LOOKUP[sitk_type]
    
    
# Math
    
    
def compute_coarse_parameters(in_dims, in_spacing, out_spacing, reduce_level):
    '''
    '''
    
    reduce_factor = pow(2, reduce_level)
    fradius = np.divide(out_spacing, in_spacing) / 2.0 / reduce_factor
    
    coarse_grid_radius = np.round(fradius)
    coarse_grid_size = (coarse_grid_radius * 2 + 1) * reduce_factor
    
    coarse_grid_spacing = np.multiply(in_spacing, coarse_grid_size)
    coarse_grid_dims = np.ceil( np.divide(in_dims, coarse_grid_size) ).astype(int)
    
    return coarse_grid_dims, coarse_grid_spacing, coarse_grid_radius
    
    
def block_apply(in_image, out_shape, dtype, blocks, fn):
    '''
    '''
    
    out_image = np.zeros(out_shape, dtype=dtype)
    
    for ii, row_block in enumerate(blocks[0]):
        for jj, col_block in enumerate(blocks[1]):
        
            out_image[ii, jj] = fn(in_image[row_block[0]:row_block[1], 
                                            col_block[0]:col_block[1]])
                                            
    return out_image
    
    
def grid_image_blocks(in_shape, in_spacing, out_spacing):
    '''
    '''

    blocks = []
    out_shape = []
    for dim in range(len(in_shape)):
        in_px_centers = np.arange(in_spacing[dim]*0.5,
                                  in_shape[dim]*in_spacing[dim],
                                  in_spacing[dim])


        out_px_edges = np.arange(out_spacing[dim],
                                 (in_shape[dim]-0.5)*in_spacing[dim],
                                 out_spacing[dim])


        dig = np.digitize(in_px_centers, out_px_edges)        

        inds = np.where(np.diff(dig)>0)[0]+1  
        inds = [0] + inds.tolist() + [in_shape[dim]]

        dim_blocks = [ (int(inds[i]), int(inds[i+1])) for i in range(len(inds)-1) ]

        out_shape.append(len(dim_blocks))
        blocks.append(dim_blocks)
        
    return blocks, out_shape
    
    
# Polygons
    
  
def rasterize_polygons(shape, scale, polys):

    canvas = np.zeros(shape, dtype=np.uint8)
    for points in polys:
    
        rpts = np.array([int(np.around(item[1] * scale[1])) for item in points])
        cpts = np.array([int(np.around(item[0] * scale[0])) for item in points])

        poly = polygon(rpts, cpts)
        canvas[poly] = 1
        
    return canvas


# Transforms
    
    
def resample_into_volume(image, transform, z, vol, dtype=sitk.sitkFloat32):
    '''
    '''

    if transform is None:
        transform = sitk.Transform()

    timage = sitk.Resample(image, transform, sitk.sitkLinear, 0.0, dtype)
    tvol = sitk.JoinSeries(timage)
    return sitk.Paste(vol, tvol, tvol.GetSize(), destinationIndex=[0,0,z])
    
    
def build_affine_transform(aff_params):
    '''
    '''

    xfm = sitk.AffineTransform(3)
    xfm.SetParameters(aff_params)
    
    return xfm
    
    
def build_composite_transform(dfmfield=None, aff_params=None):
    '''
    '''

    if dfmfield is not None and dfmfield.GetPixelIDValue() != sitk.sitkVectorFloat64:
        dfmfield = sitk.Cast(dfmfield, sitk.sitkVectorFloat64)

    if dfmfield is None and aff_params is None:
        transform = sitk.Transform()
    elif dfmfield is not None and aff_params is None:
        transform = sitk.DisplacementFieldTransform(dfmfield)
    elif dfmfield is None and aff_params is not None:
        transform = build_affine_transform(aff_params)
    elif dfmfield is not None and aff_params is not None:
        dfmxfm = sitk.DisplacementFieldTransform(dfmfield)
        affxfm = build_affine_transform(aff_params)

        transform = sitk.Transform()
        transform.AddTransform(affxfm)
        transform.AddTransform(dfmxfm)

    return transform
    
    
def resample_volume(volume, dims, spacing, interpolator=None, transform=None):
    '''
    '''

    if transform is None:
        transform = sitk.Transform() 
    if interpolator is None:
        interpolator = sitk.sitkLinear
    
    ref = new_image(dims, spacing, sitk.sitkFloat32, False)
    return sitk.Resample(volume, ref, transform, interpolator)


def write_volume(volume, name, prefix=None, specify_resolution=None, extension='.nrrd', paths=None):

    if prefix is None:
        path = name
    else:
        path = os.path.join(prefix, name)

    if specify_resolution is not None:
        path = path + '_{0}'.format(specify_resolution)

    path = path + extension
    
    logging.info('writing {0} volume to {1}'.format(name, path))
    Manifest.safe_make_parent_dirs(path)
    volume.SetOrigin([0, 0, 0])
    sitk.WriteImage(volume, str(path), True)

    if paths is not None:
        paths.append(path)



import numpy as np
from scipy import stats
from scipy.optimize import curve_fit, fmin
from allensdk.internal.model.glif.find_spikes import align_and_cut_spikes, ALIGN_CUT_WINDOW
import logging

import matplotlib.pyplot as plt 

def calc_spike_cut_and_v_reset_via_expvar_residuals(all_current_list, 
                                                    all_voltage_list, dt, El_reference, deltaV, 
                                                    max_spike_cut_time=False,
                                                    MAKE_PLOT=False, SHOW_PLOT=False, PUBLICATION_PLOT=False, BLOCK=False):
    '''This function calculates where the spike should be cut based on explained variance.  
        The goal is to find a model where the voltage after a spike maximally explains the
        voltage before a spike.  This will also specify the voltage reset rule
        inputs: 
            spike_determination_method:  string specifing the method used to find threshold
            all_current_list: list of current (list of current traces injected into neuron)
            all_voltage_list: list of voltages (list of voltage trace)
       The change is that if the slope is greater than one or intercept is greater than zero it forces it.  
       Regardless of required force the residuals are used.                 
    '''
   
    #--find the region of the spike needed for calculation of explained variance
    (temp_v_spike_shape_list, all_i_spike_shape_list, all_thresholdInd, waveIndOfFirstSpikes, spikeFromWhichSweep) \
                            = align_and_cut_spikes(all_voltage_list, all_current_list, dt)
    
    #--At this point it is unclear how this calculation should be done.  
    #--the slope should be fine no matter what, but the intercept dependency
    #--will depend on the El, and deltaV    
    
    #--change reference 
    all_v_spike_shape_list=[shape-El_reference-deltaV for shape in temp_v_spike_shape_list]

    # --setting limits to find explained variance
    if max_spike_cut_time and max_spike_cut_time < .010:
        expVarIndRangeAfterSpike = range(int(.001 / dt), int(max_spike_cut_time / dt))  #NOTE: THIS IS USED IN REFERENCE TO SPIKE TIME

    else:
        expVarIndRangeAfterSpike = range(int(.001 / dt), int(.010 / dt))  #NOTE: THIS IS USED IN REFERENCE TO SPIKE TIME
    vectorIndex_of_max_explained_var = expVarIndRangeAfterSpike[0]  # this is just here for the title of the plot
    list_of_endPointArrays = []  # this should end up a list of numpy arrays where each numpy array contains the indices of the v_spike_shape_list that are a certain time after the threshold
    for ii in expVarIndRangeAfterSpike:
        list_of_endPointArrays.append(np.array(all_thresholdInd) + ii)
        
    def line_force_slope_to_1(x,c):
        return x+c
    
    def line_force_int_to_0(x, m):  #TODO: CHANGE THIS TO REST TOD DISCONNECT EVERYTHING.
        return m*x

#    HERE YOU GET THE SLOPE AND INTERCEPT AT EACH POINT
    linRegress_error_4_each_time_end = []
    slope_at_each_time_end=[]
    intercept_at_each_time_end=[]
    varData_4_each_time_end = []
    varModel_4_each_time_end = []
    chi2 = []
    sum_residuals_4_each_time_end=[]
    xdata = np.array([v[all_thresholdInd[ii]] for ii, v in enumerate(all_v_spike_shape_list)])
    var_of_Vdata_beforeSpike = np.var(xdata)
    for jj, vectorOfIndAcrossWaves in enumerate(list_of_endPointArrays):  # these indices should be in terms of the spike waveforms
#        print 'jj', jj
        # TODO: Teeter get rid of the nonblipness
        v_at_specificEndPoint = [all_v_spike_shape_list[ii][index] for ii, index in enumerate(vectorOfIndAcrossWaves)]  # this is calculating variance at certain time points
        # --currently the model of voltage reset is a linear regression between voltage before the spike and the voltage after the spike but it could be more complicated (for example as a function of current) 
        ydata = np.array(v_at_specificEndPoint)  # this is the voltage at the specified end point
        slope, intercept, r_value, p_value, std_err = stats.linregress(xdata, ydata)

#        print slope, intercept, r_value, p_value, std_err
    
#        if slope>1.0:
#            logging.warning('linear regression slope is bigger than one: forcing slope to 1 and refitting intercept.')
#            slope=1.0
#            (intercept, nothing)=curve_fit(line_force_slope_to_1, xdata, ydata)
#            #print "NEW INTERCEPT:", intercept
#            if intercept>0.0:
#                #warnings.warn('/t ... and intercept is bigger than zero: forcing intercept to 0')
#                intercept=0.0
#            
#        if intercept>0.0:
#            logging.warning('Intercept is bigger than zero: forcing intercept to 0 and refitting slope.')
#            intercept=0.0                        
#            (slope, nothing)=curve_fit(line_force_int_to_0,xdata, ydata)  
#            #print "NEW SLOPE: ", slope
#            if slope>1.0:
#                logging.warning('/t ... and linear regression slope is bigger than one: forcing slope to 1.')
#                slope=1.0
              
        slope_at_each_time_end.append(slope)
        intercept_at_each_time_end.append(intercept)
        ymodel = slope * xdata + intercept
        residuals = ydata - ymodel
        sum_residuals=sum(abs(residuals))
        sum_residuals_4_each_time_end.append(sum_residuals)
        chi2.append(np.var(residuals))  # how well the model describes the data
        linRegress_error_4_each_time_end.append(std_err)
        varData_4_each_time_end.append(np.var(v_at_specificEndPoint))
        varModel_4_each_time_end.append(np.var(ymodel))
        
    # --these will line up with how many arrays there are in the list
    vectorIndex_of_min_sum_residuals = sum_residuals_4_each_time_end.index(min(sum_residuals_4_each_time_end))

    #----NOTE THIS ISNT ACTUALLY CALCULATING EXPLAINED VARIANCE!!!!!!!!!!!!!!!!!!
    vectorIndex_of_max_explained_var=vectorIndex_of_min_sum_residuals
    
    
    all_v_spike_init_list = [v[all_thresholdInd[ii]] for ii, v in enumerate(all_v_spike_shape_list)]
# USE THIS WHEN MUTIPLE VECTORS all_v_at_min_expVar_list=[v[list_of_endPointArrays[vectorIndex_of_max_explained_var][ii]] for ii, v in enumerate(all_v_spike_shape_list)] 
    all_v_at_min_expVar_list = [v[list_of_endPointArrays[vectorIndex_of_max_explained_var][ii]] for ii, v in enumerate(all_v_spike_shape_list)] 
    time_at_minExpVar=list_of_endPointArrays[vectorIndex_of_max_explained_var]*dt
    if MAKE_PLOT:
        truncatedTime = np.arange(0, len(all_v_spike_shape_list[0])) * dt
        plt.figure(figsize=(20, 10))
        for ii in range(0, len(all_v_spike_shape_list)):
            plt.subplot(2,1,1)
            plt.plot(truncatedTime, temp_v_spike_shape_list[ii])
            # plt.plot(truncatedTime[aligned_peakInd[ii]],spikewave[aligned_peakInd[ii]], '.k'
            plt.plot(truncatedTime[all_thresholdInd[ii]], temp_v_spike_shape_list[ii][all_thresholdInd[ii]], '*k')
            plt.title('Non adusted spikes')
            
            plt.subplot(2,1,2)
            plt.plot(truncatedTime, all_v_spike_shape_list[ii])
            plt.plot(time_at_minExpVar, all_v_at_min_expVar_list, '*k')
            plt.xlabel('time (s)', fontsize=20)
            plt.ylabel('voltage (mV)', fontsize=20)
            plt.title("Adjusted spikes (RP=%.3g, deltaV=%.3g)" % (El_reference,deltaV))
        
        if PUBLICATION_PLOT:
            truncatedTime = np.arange(0, len(all_v_spike_shape_list[0])) * dt
            plt.figure(figsize=(20, 5))
            for ii in range(0, len(all_v_spike_shape_list)):
                plt.plot(truncatedTime*1000, temp_v_spike_shape_list[ii]*1e3, lw=2)
                # plt.plot(truncatedTime[aligned_peakInd[ii]],spikewave[aligned_peakInd[ii]], '.k'
                plt.plot(truncatedTime[all_thresholdInd[ii]]*1000, temp_v_spike_shape_list[ii][all_thresholdInd[ii]]*1e3, '.k', ms=10)
#                plt.title('Spike Cutting', fontsize=20)
#                plt.subplot(2,1,2)
#                plt.plot(truncatedTime, all_v_spike_shape_list[ii])
                plt.plot(time_at_minExpVar*1000, (np.array(all_v_at_min_expVar_list)+El_reference+deltaV)*1.e3, '.k', ms=10)
                plt.xlabel('Time (ms)', fontsize=16)
                plt.ylabel('Voltage (mV)', fontsize=16)
                plt.xlim([0,12])
                plt.tight_layout()
#                plt.title("Adjusted spikes (RP=%.3g, deltaV=%.3g)" % (El_reference,deltaV))  
            
        if SHOW_PLOT:
            plt.show(block=BLOCK)    


#    indNotExcluded_In_regress=list(np.setdiff1d(np.array([theInd for theInd in spikeIndDict['nonblip']]), np.array(waveIndOfFirstSpikes)))
#    something is wrong with all_v_at_min_expVar_list--look at the difference between starting at .003 and .005 after thresh
    if MAKE_PLOT:
        plt.figure(figsize=(20, 10))
        plt.plot(all_v_spike_init_list, all_v_at_min_expVar_list, 'b.', ms=16, label='noise')  # list of voltage traces for blip
        plt.xlabel('voltage at spike initiation (V)', fontsize=20)
        plt.ylabel('voltage after spike (V)', fontsize=20)
#                    plt.title(cellTitle, fontsize=20)

    slope_at_min_expVar_list, intercept_at_min_expVar_list, r_value_at_min_expVar_list, p_value_at_min_expVar_list, std_err_at_min_expVar_list = \
        stats.linregress(np.array(all_v_spike_init_list), np.array(all_v_at_min_expVar_list))

    print 'mean of voltage before spike', np.mean(all_v_spike_init_list)
    print 'mean of voltage after spike', np.mean(all_v_at_min_expVar_list)
    
    
    spike_cut_length= (list_of_endPointArrays[vectorIndex_of_max_explained_var][0])-int(ALIGN_CUT_WINDOW[0]/dt) #note this is dangerous if they arent' all at the same ind

    
    if MAKE_PLOT:
        xlim = np.array([min(all_v_spike_init_list), max(all_v_spike_init_list)])
        plotLineRegress1(slope_at_min_expVar_list, intercept_at_min_expVar_list, r_value_at_min_expVar_list, xlim)
        plt.legend(loc=2, fontsize=20)


    if MAKE_PLOT:
        xlim = np.array([min(all_v_spike_init_list), max(all_v_spike_init_list)])
        plotLineRegressRed(slope_at_each_time_end[vectorIndex_of_max_explained_var], intercept_at_each_time_end[vectorIndex_of_max_explained_var], np.NAN, xlim)
        plt.legend(loc=2, fontsize=20)
        if SHOW_PLOT:
            plt.show(block=BLOCK)                   

    if PUBLICATION_PLOT:
        
        plt.figure(figsize=(7, 5))
        plt.plot(np.array(all_v_spike_init_list)*1e3, np.array(all_v_at_min_expVar_list)*1e3, 'b.', ms=16)  # list of voltage traces for blip
        plt.xlabel('Voltage at spike initiation (mV)', fontsize=16)
        plt.ylabel('Voltage after spike (mV)', fontsize=16)
#        plt.title('Voltage reset rules', fontsize=20)
        xlim = np.array([min(all_v_spike_init_list), max(all_v_spike_init_list)])
        
        def plot_hack(slope, intercept, r,xlim):
            y=slope*xlim+intercept
            plt.plot(xlim, y, '-k', lw=4)# label='slope='+"%.2f"%slope+', intercept='+"%.3f"%intercept)
            
        plot_hack(slope_at_min_expVar_list, intercept_at_min_expVar_list*1e3, r_value_at_min_expVar_list, xlim*1e3)
        plt.legend(loc=2, fontsize=16)
        plt.tight_layout()
        plt.show(block=BLOCK)

    #TODO:  Corinne look to see if these were calculated with zeroed out El if not does is matter?
    if isinstance(slope_at_min_expVar_list, np.ndarray):
        slope_at_min_expVar_list=float(slope_at_min_expVar_list[0])
    if isinstance(intercept_at_min_expVar_list, np.ndarray):
        intercept_at_min_expVar_list=float(intercept_at_min_expVar_list[0])

    if type(intercept_at_min_expVar_list)==list or type(intercept_at_min_expVar_list)==np.ndarray:
        intercept_at_min_expVar_list=intercept_at_min_expVar_list[0]

    return spike_cut_length, slope_at_min_expVar_list, intercept_at_min_expVar_list

def plotLineRegress1(slope, intercept, r,xlim):
    y=slope*xlim+intercept
    print 'slope=', slope, 'intercept=', intercept, 'xlim', xlim
    plt.plot(xlim, y, '-k', lw=4, label='slope='+"%.2f"%slope+', intercept='+"%.3f"%intercept+', r='+"%.2f"%r)

def plotLineRegressRed(slope, intercept, r,xlim):
    y=slope*xlim+intercept
    print 'slope=', slope, 'intercept=', intercept, 'xlim', xlim
    plt.plot(xlim, y, '-r', lw=4, label='slope='+"%.2f"%slope+', intercept='+"%.3f"%intercept+', r='+"%.2f"%r)



import argparse
import allensdk.internal.core.lims_utilities as lu
import glob
import time
import shutil
import logging
from allensdk.config.manifest import Manifest
from allensdk.internal.brain_observatory.itracker import iTracker
from allensdk.internal.brain_observatory.frame_stream import FfmpegInputStream, FfmpegOutputStream
import h5py
import ast
import sys
import numpy as np

DEFAULT_THRESHOLD_FACTOR = 1.6

if sys.platform=='linux2':
    FFMPEG_BIN = "/shared/utils.x86_64/ffmpeg/bin/ffmpeg"
elif sys.platform=='darwin':
    FFMPEG_BIN = "/usr/local/bin/ffmpeg"

def compute_bounding_box(points):
    if not points:
        return None
    points = np.array(points)
    return [ points[:,0].min(), points[:,0].max(),
             points[:,1].min(), points[:,1].max() ]

def get_polygon(experiment_id, group_name):
    query = """
select ago.* from avg_graphic_objects ago
join avg_graphic_objects pago on pago.id = ago.parent_id
join avg_group_labels agl on pago.group_label_id = agl.id
join sub_images si on si.id = pago.sub_image_id
join specimens sp on sp.id = si.specimen_id
join ophys_sessions os on os.specimen_id = sp.id
where agl.name = '%s'
and os.id = %d
""" % (group_name, experiment_id)

    try:
        path = np.array([ int(v) for v in lu.query(query)[0]['path'].split(',') ])
    except KeyError as e:
        return []
    except IndexError as e:
        return []

    points = path.reshape((len(path)/2, 2))

    return points

def get_experiment_info(experiment_id):
    logging.info("Downloading paths/metadata for experiment ID: %d", experiment_id)
    query = "select storage_directory, id from ophys_sessions where id = "+str(experiment_id)

    storage_directory = lu.query(query)[0]['storage_directory']
    logging.info("\tStorage directory: %s", storage_directory)

    movie_file = glob.glob(storage_directory+'*video-1.avi')[0]
    metadata_file = glob.glob(storage_directory+'*video-1.h5')[0]

    cr_points = get_polygon(experiment_id, 'Corneal Reflection Bounding Box')
    pupil_points = get_polygon(experiment_id, 'Pupil Bounding Box')

    logging.info("\tmovie file: %s", movie_file)
    logging.info("\tmetadata file: %s", metadata_file)

    return dict(movie_file=movie_file,
                metadata_file=metadata_file,
                corneal_reflection_points=cr_points,
                pupil_points=pupil_points)

def get_movie_shape_from_metadata(metadata_file):
    with h5py.File(metadata_file, "r") as f:
        metadata_str = f["video_metadata"].value
        metadata = ast.literal_eval(metadata_str)

    # assuming 3 channels
    # movie_shape = (metadata['frames'], metadata['height'], metadata['width'], 3)
    # in the metadata file from lims, the 'width' and 'height' variables are swapped,
    # hopefully this is the same for every single experiment.
    movie_shape = (metadata['frames'], metadata['width'], metadata['height'], 3)
    logging.info("movie_shape from metadata_file = %s", str(movie_shape))

    return movie_shape

def run_itracker(movie_file, output_directory,
                 output_frames=False,
                 output_annotation_frames=False,
                 output_annotated_movie=True,
                 output_annotated_movie_block_size=1,
                 estimate_bbox=False,
                 num_frames=None,
                 output_QC=True,
                 image_type='png',
                 cache_input_frames=False,
                 input_block_size=1,
                 metadata_file=None,
                 movie_shape=None,
                 **kwargs):

    if output_directory is not None:
        Manifest.safe_mkdir(output_directory)

    assert(metadata_file is not None and movie_shape is not None, "Must provide either metadata_file or movie_shape")

    if metadata_file:
        movie_shape = get_movie_shape_from_metadata(metadata_file)
       
    frame_shape = movie_shape[1:]

    if num_frames is None:
        num_frames = movie_shape[0]


    input_stream = FfmpegInputStream(movie_file, frame_shape,
                                     ffmpeg_bin=FFMPEG_BIN,
                                     num_frames=num_frames,
                                     cache_frames=cache_input_frames,
                                     block_size=input_block_size)

    movie_output_stream = FfmpegOutputStream(frame_shape, 
                                             block_size=output_annotated_movie_block_size,
                                             ffmpeg_bin=FFMPEG_BIN) if output_annotated_movie else None

    itracker = iTracker(output_directory, input_stream=input_stream,
                        im_shape=(movie_shape[1], movie_shape[2]),
                        num_frames=num_frames,
                        **kwargs)

    # open this early to avoid duplicating massive memory
    movie_output_stream.open(itracker.annotated_movie_file)

    itracker.set_movie(movie_file)

    itracker.mean_frame = itracker.compute_mean_frame()

    if estimate_bbox:
        bbox_pupil, bbox_cr = itracker.estimate_bbox_from_mean_frame()


    itracker.process_movie(movie_output_stream=movie_output_stream,
                           output_frames=output_frames,
                           output_annotation_frames=output_annotation_frames)

    if output_QC:
        itracker.output_QC(image_type=image_type)

    return itracker


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--experiment_id', default=None, type=int)
    parser.add_argument('--movie_file', default=None)
    parser.add_argument('--metadata_file', default=None)
    parser.add_argument('--output_directory', default='.')
    parser.add_argument('--estimate_bbox', action='store_true')
    parser.add_argument('--num_frames', default=None, type=int)
    parser.add_argument('--threshold_factor', default=DEFAULT_THRESHOLD_FACTOR)
    parser.add_argument('--log_level', default=logging.DEBUG)
    args = parser.parse_args()

    logging.getLogger().setLevel(args.log_level)

    data = dict(
        threshold_factor=args.threshold_factor,
        output_directory=args.output_directory,
        num_frames=args.num_frames,
        estimate_bbox=args.estimate_bbox
        )

    if args.experiment_id:
        info = get_experiment_info(args.experiment_id)

        data['movie_file'] = info['movie_file']
        data['metadata_file'] = info['metadata_file']

        if info.get('pupil_points', None):
            data['bbox_pupil'] = compute_bounding_box(info['pupil_points'])
        if info.get('corneal_reflection_points', None):
            data['bbox_cr'] = compute_bounding_box(info['corneal_reflection_points'])
    else:
        data['movie_file'] = args.movie_file
        data['metadata_file'] = args.metdata_file

    run_itracker(**data)

if __name__ == "__main__": main()



import numpy as np
import os
import allensdk.internal.model.biophysical.passive_fitting.neuron_utils as neuron_utils

import allensdk.core.json_utilities as json_utilities
from allensdk.model.biophys_sim.config import Config

# Load the morphology

BASEDIR = os.path.dirname(__file__)

@neuron_utils.read_neuron_fit_stdout
def neuron_passive_fit2(up_data, down_data, swc_path, limit):
    h = neuron_utils.get_h()
    h.load_file("stdgui.hoc")
    h.load_file("import3d.hoc")
    neuron_utils.load_morphology(swc_path)

    for sec in h.allsec():
        sec.insert('pas')
        for seg in sec:
            seg.pas.e = 0

    h.load_file(os.path.join(BASEDIR, "passive", "fixnseg.hoc"))
    h.load_file(os.path.join(BASEDIR, "passive", "iclamp.ses"))
    h.load_file(os.path.join(BASEDIR, "passive", "params.hoc"))
    h.load_file(os.path.join(BASEDIR, "passive", "mrf2.ses"))

    h.v_init = 0
    h.tstop = 100

    fit_start = 4.0025

    v_rec = h.Vector()
    t_rec = h.Vector()
    v_rec.record(h.soma[0](0.5)._ref_v)
    t_rec.record(h._ref_t)

    mrf = h.MulRunFitter[0]
    gen0 = mrf.p.pf.generatorlist.object(0)
    gen0.toggle()
    fit0 = gen0.gen.fitnesslist.object(0)

    up_t = h.Vector(up_data[:, 0])
    up_v = h.Vector(up_data[:, 1])
    fit0.set_data(up_t, up_v)
    fit0.boundary.x[0] = fit_start
    fit0.boundary.x[1] = limit
    fit0.set_w()

    gen1 = mrf.p.pf.generatorlist.object(1)
    gen1.toggle()
    fit1 = gen1.gen.fitnesslist.object(0)

    down_t = h.Vector(down_data[:, 0])
    down_v = h.Vector(down_data[:, 1])
    fit1.set_data(down_t, down_v)
    fit1.boundary.x[0] = fit_start
    fit1.boundary.x[1] = limit
    fit1.set_w()

    minerr = 1e12
    for _ in range(3):
        # Need to re-initialize the internal MRF variables, not top-level proxies
        # for randomize() to work
        mrf.p.pf.parmlist.object(0).val = 1
        mrf.p.pf.parmlist.object(1).val = 10000
        mrf.randomize()
        mrf.prun()
        if mrf.opt.minerr < minerr:
            fit_Ri = h.Ri
            fit_Cm = h.Cm
            fit_Rm = h.Rm
            minerr = mrf.opt.minerr

    h.region_areas()
    return {
        'Ri': fit_Ri,
        'Cm': fit_Cm,
        'Rm': fit_Rm,
        'err': minerr
        }

def main():
    import sys
    
    manifest_path = sys.argv[-1]
    limit = float(sys.argv[-2])
    os.chdir(os.path.dirname(manifest_path))
    app_config = Config()
    description = app_config.load(manifest_path)
    
    upfile = description.manifest.get_path('upfile')
    up_data =  np.loadtxt(upfile)
    downfile = description.manifest.get_path('downfile')
    down_data = np.loadtxt(downfile)
    swc_path = description.manifest.get_path('MORPHOLOGY')
    output_file = description.manifest.get_path('fit_2_file')
    
    data = neuron_passive_fit2(up_data, down_data, swc_path, limit)
    
    json_utilities.write(output_file, data)

if __name__ == "__main__": main()



import numpy as np
import copy
from scipy.interpolate import interp1d
from scipy.optimize import curve_fit
import matplotlib.pyplot as plt
import logging
THRESH_PCT_MULTIBLIP = 0.05

from allensdk.model.glif.glif_neuron_methods import spike_component_of_threshold_exact
from allensdk.internal.model.glif.find_spikes import find_spikes_ssq_list

def calc_spike_component_of_threshold_from_multiblip(multi_SS, dt, dv_cutoff, thresh_frac,
                                                     MAKE_PLOT=False, SHOW_PLOT=False, BLOCK=False, PUBLICATION_PLOT=False):
    '''Calculate the spike components of the threshold by fitting a decaying exponential function to data to threshold versus time 
    since last spike in the multiblip data. The exponential is forced to decay to the local th_inf (calculated as the mean all of the 
    threshold values of the first spikes in each individual triblip stimulus).  For each multiblip stimulus in a stimulus set if there
    is more than one spike the difference in voltages from the first and second spike are plotted versus the separation in time.  Note that 
    this algorithm should only be implemented on multiblips sweeps where the neuron spike on the first and second blip.  Since there is 
    no easy way to do this, this erroneous data should not be provided to this algorithm (i.e is should be visually checked and eliminated 
    the preprocessor should hold back this data manually for now.)  
    
    #TODO: check to see if this is still true. Notes: The standard SDK spike detection algorithm does not work with the multiblip stimulus 
    due to artifacts when the stimulus turns on and off. Please see the find_multiblip_spikes module for more information.
    
    Input:
    
    multi_SS: dictionary
        contains multiblip information such as current and stimulus
    dt: float
        time step in seconds
    
    Returns:
    
    const_to_add_to_thresh_for_reset: float
        amplitude of the exponential fit otherwise known as a_spike.  Note that this is without any spike cutting
    decay_const: float
        decay constant of exponential. Note the function fit is a negative exponential which will mean this value will 
        either have to be negated when it is used or the functions used will have to have to include the negative.
    thresh_inf: float
    
    '''    
    multi_SS_v=multi_SS['voltage']
    multi_SS_i=multi_SS['current']

    # --get indicies of spikes
    spike_ind, _=find_spikes_ssq_list(multi_SS_v, dt, dv_cutoff, thresh_frac)
#    spike_ind=find_multiblip_spikes(multi_SS_i, multi_SS_v, dt) can depricate find_multiblip_spikes


    # eliminate spurious spikes that may exist
    spike_lt=[np.where(SI<int(2.0/dt))[0] for SI in spike_ind]
    if len(np.concatenate(spike_lt))>0:
        logging.warning('there is a spike before the stimulus in the multiblip')
    spike_ind=[np.delete(SI,ind)for SI, ind in zip(spike_ind, spike_lt)]
    spike_gt=[np.where(SI>int(3.0/dt))[0] for SI in spike_ind]
    if len(np.concatenate(spike_gt))>0:
        logging.warning('there is a spike after the stimulus in the multiblip')    
    spike_ind=[np.delete(SI,ind)for SI, ind in zip(spike_ind, spike_gt)]
    
    # intialize output lists
    time_previous_spike=[]
    threshold=[]
    thresh_first_spike=[]  #will set constant to this

    if MAKE_PLOT:
        plt.figure(figsize=(20,24))
        
    # Loop though each tri blip stimulus in muliblip stimulus
    for k in range(0, len(multi_SS_v)):
        thresh=[multi_SS_v[k][j] for j in spike_ind[k]] # voltage at all spikes in a single tri blip
        if thresh!=[] and len(thresh)>1:# there needs to be more than one spike so that we can find the time difference
            thresh_first_spike.append(thresh[0]) #Note that this finds the first spike (it might not be at the first stimulus blip)
            threshold.append(thresh[1])
            time_previous_spike.append((spike_ind[k][1]-spike_ind[k][0])*dt)
#            Old way when looked at all the spikes instead of just the first two (can be depricated; just here for record keeping)
#            threshold.append(thresh[1:])              
#            time_before_temp=[]
#            for j in range(1,len(thresh)):
#                time_before_temp.append((spike_ind[k][j]-spike_ind[k][j-1])*dt)
#            #for each spike calculate the time from the previous spike   
#            time_previous_spike.append(time_before_temp)    
        if MAKE_PLOT:
            plt.subplot(len(multi_SS_v)+1,1,1)
            plt.plot(np.arange(0, len(multi_SS_i[k]))*dt, multi_SS_i[k]*1e12, lw=2)
            plt.ylabel('current (pA)', fontsize=16)
            plt.xlim([2., 2.12])
            plt.title('Triple Short Square', fontsize=20)
            plt.subplot(len(multi_SS_v)+1,1,k+2)
            plt.plot(np.arange(0, len(multi_SS_v[k]))*dt, multi_SS_v[k], lw=2)
            plt.plot(spike_ind[k]*dt, thresh, '.k', ms=16)
            plt.xlim([2., 2.12])
            
    if MAKE_PLOT:
            plt.ylabel('voltage (V)', fontsize=16)
            plt.xlabel('time (s)', fontsize=16)
    
    if SHOW_PLOT:        
        plt.show(block=False)

    # put numbers into one vector for fitting of exponential function
    thresh_inf=np.mean(thresh_first_spike)  #note this threshold infinity isnt the one coming from single blip
    try: #this try here because sometimes even though have the traces there isnt more than one trace with two spikes
#--these two lines no longer needed because all single values now (depricate with lines up above) instead converst them to arrays
#        threshold=np.concatenate(threshold)
#        time_previous_spike=np.concatenate(time_previous_spike)  #note that this will have nans in it
        threshold=np.array(threshold)
        time_previous_spike=np.array(time_previous_spike)  #note that this will have nans in it
   
        if MAKE_PLOT:
            plt.figure()
            plt.plot(time_previous_spike, threshold, '.k', ms=16)
            plt.ylabel('threshold (mV)')
            plt.xlabel('time since last spike (s)')
    
        # calculate values of exponential function both if force function to local threshold infinity and not forcing to a value
        # (not forcing to a value seems less valid unless a bunch of points are added corresponding to the threshold of the
        # first spike at time equal infinity (because the first spike is a spike that happens where the spike before it was an
        # infinite time away)).  Therefore, the values that are obtained from forcing are the ones that are used.
        p0_force=[.002, -100.]
        p0_fit=[.002, -100., thresh_inf]
        
        #TODO: THIS WOULD BE BETTER IF IT CALLED THE ACTUAL FUNCTION IN THE NEURON METHODS THAT WAY THEY WOULD HAVE TO BE THE SAME
        (popt_force, pcov_force)= curve_fit(exp_force_c, (time_previous_spike, thresh_inf), threshold, p0=p0_force, maxfev=100000)
        (popt_fit, pcov_fit)= curve_fit(exp_fit_c, time_previous_spike, threshold, p0=p0_fit, maxfev=100000)

        # viewing fit functions
        time_previous_spike.sort() #since time is not in order, making new time vector so that obtained fit curve can be plotted
        fit_force=exp_force_c((time_previous_spike, thresh_inf), popt_force[0], popt_force[1])
        fit_fit=exp_fit_c(time_previous_spike, popt_fit[0], popt_fit[1], popt_fit[2])
        if MAKE_PLOT:
            plt.plot(time_previous_spike, fit_force, 'r', lw=4, label="exp fit (force const to thesh first spike)\n  k=%.3g, amp=%.3g" % (popt_force[1], popt_force[0]))
            plt.plot(time_previous_spike, fit_fit, 'b', lw=4, label="exp fit (fit constant)\n  k=%.3g, amp=%.3g" % (popt_fit[1], popt_fit[0]))
            plt.legend()
            if SHOW_PLOT:        
                plt.show(block=False)
    
            if PUBLICATION_PLOT:       
                plt.figure(figsize=[14, 5])
                ax1=plt.subplot2grid((2, 2), (0,0))
                ax2=plt.subplot2grid((2, 2), (1,0))
                ax3=plt.subplot2grid((2, 2), (0,1), rowspan=2)
                for k in range(0, len(multi_SS_v)):
                    thresh=[multi_SS_v[k][j] for j in spike_ind[k]]
    
                    ax1.plot(np.arange(0, len(multi_SS_i[k]))*dt, multi_SS_i[k]*1.e12, lw=2)
                    ax1.set_ylabel('Current (pA)', fontsize=16)
                    ax1.set_xlim([2., 2.12])
                    ax1.axes.xaxis.set_ticklabels([])
                    #ax1.set_title('Triple Short Square', fontsize=20)
    
                    ax2.plot(np.arange(0, len(multi_SS_v[k]))*dt, multi_SS_v[k]*1.e3, lw=2)
                    ax2.plot(spike_ind[k]*dt, np.array(thresh)*1.e3, '.k', ms=16)
                    ax2.set_ylabel('Voltage (mV)', fontsize=16)
                    ax2.set_xlabel('Time (ms)', fontsize=16)
                    ax2.set_xlim([2., 2.12])
                    
    
                ax3.plot(time_previous_spike, np.array(threshold)*1.e3, '.k', ms=16)
                ax3.set_ylabel('Threshold (mV)', fontsize=16)
                ax3.set_xlabel('Time since last spike (s)', fontsize=16)
#                ax3.set_title('Spiking component of threshold', fontsize=20)
                ax3.plot(time_previous_spike, fit_force*1.e3, 'r', lw=4)#, label="exp fit: k=%.3g, amp=%.3g" % (popt_force[1], popt_force[0]))
                ax3.legend() 
                plt.tight_layout()
                plt.show()
        
        const_to_add_to_thresh_for_reset=popt_force[0]
        decay_const=popt_force[1]
        
        if decay_const >0:
            logging.critical('This neuron has an increasing decay value for the spike component of the threshold')
        if const_to_add_to_thresh_for_reset<0:
            logging.critical('This neuron has a negative amplitude for the spike component of the threshold') 

        #if the decay constant is positive, or the amplitute is negative set the amplitude to 0 so that there 
        #will be no spike component of the threshold
        if decay_const >0 or const_to_add_to_thresh_for_reset < 0:
            const_to_add_to_thresh_for_reset=0
            decay=-1.0  #note that this number doesnt matter since the amplitude is set to zero
            
        # This decay constant was originally forced to be positive (i.e. decay_const=abs(popt_force[1]))
        # and then it is negated everywhere it is utilized elsewhere in the code.  Now things are forced in 
        # a different way above.  However the decay constant still needs to be negated here for use in the 
        # rest of the code. 
        decay_const=-decay_const
    

    except Exception, e:
        logging.error(e.message)
        const_to_add_to_thresh_for_reset=None 
        decay_const=None
        
    return const_to_add_to_thresh_for_reset, decay_const, thresh_inf

def fit_avoltage_bvoltage(x, v_trace_list, El_list, spike_cut_length, all_spikeInd_list, th_inf, dt, a_spike, 
                     b_spike, fake=False):
    '''This is a version of fit_avoltage_bvoltage_debug that does not require the th_trace, 
    v_component_of_thresh_trace, and spike_component_of_thresh_trace needed for debugging. A
    test should be run to make sure the same output comes out from this and the debug function
    
    This function returns the squared error for the difference between the 'known' voltage
    component of the threshold obtained from the biological neuron and the voltage component 
    of the threshold of the model obtained with the input parameters (so that the minimum can be 
    searched for via fmin). The overall threshold is the sum of threshold infinity the spike component
    of the threshold and the voltage component of the threshold.  Therefore threshold infinity and 
    the spike component of the threshold must be subtracted from the threshold of the neuron in order
    to isolate the voltage component of the threshold.  In the evaluation of the model the actual
    voltage of the neuron is used so that any errors in the other components of the model will not 
    influence the fits here (for example, if a afterspike current was estimated incorrectly)
  
    Notes:
    * The spike component of the threshold is subtracted from the 
        voltage which means that the voltage component of the threshold should only be added to rules.
    * b_spike was fit using a negative value in the function therefore the negative is placed in the 
        equation. 
    * values in this function are in 'real' voltage as opposed to voltage
        relative to resting potential. 
    * current injection during the spike is not taken into account.  This seems reasonable as the 
        ion channels are open during this time and injected current may not greatly influence the neuron.
    
    x: numpy array
        x[0]=a_voltage input, x[1] is b_voltage_input, x[2] is th_inf
    v_trace_list: list of numpy arrays
        voltage traces (v_trace, El, and th_inf must be in the same frame of reference)
    El_list: list of floats
        reversal potential (v_trace, El, and th_inf must be in the same frame of reference)
    spike_cut_length: int
        number of indicies removed after initiation of a spike
    all_spikeInd_list: list of numpy arrays
        indicies of spike trains 
    th_inf: float
        threshold infinity (v_trace, El, and th_inf must be in the same frame of reference)
    dt: float 
        size of time step (SI units)
    a_spike: float
        amplitude of spike component of threshold.
    b_spike: float
        decay constant in spike component of the threshold
    fake: Boolean
        if True makes uses the voltage value of spike step-1 because there is not a voltage value at the spike
        step because it is set to nan in the simulator.
    '''
    a_voltage=x[0]
    b_voltage=x[1]

    total_err=0
    for v_trace, El, all_spikeInd in zip(v_trace_list, El_list, all_spikeInd_list):
        # Calculate values along the whole trace and then take the values at the spike ind
        internal_sp_comp_array=np.zeros(all_spikeInd[0]+spike_cut_length) 
        left_over=0
        #vector of spike component of of the threshold from each spike and previous spikes; note at first spike there is no spike component of threshold so initialized at zero   
        #Note that care has to be taken here to get make sure the right amount of decay is left over
        for spike_number in range(1,len(all_spikeInd)):      #skipping first spike since no residual spike component of threshold
            integration_length=all_spikeInd[spike_number]-all_spikeInd[spike_number-1]+1 #this is the amount of time that needs to be integrated over note that it is one longer than the interval because of the last value is added to the aspike for the next ISI
            local_spike_comp_of_threshold=spike_component_of_threshold_exact(a_spike+left_over, b_spike, np.arange(integration_length)*dt)
            internal_sp_comp_array=np.append(internal_sp_comp_array, local_spike_comp_of_threshold[:-1])
            left_over=local_spike_comp_of_threshold[-1]
        
        # Compute voltage component of threshold at biological spike (subtract th_inf and spike component of threshold
        # from biological voltage values at spike initiation)
        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        #     NOTE THAT THERE IS AN ISSUE HERE USING FAKE DATA.  THE -1 IS HERE BECAUSE THE NEURON CROSSES THRESHOLD SOMETIME BETWEEN TWO INDICIES.
        #     FOR THE FAKE DATA THE TIME OF THE SPIKE (THE POINT FOLLOWING WHEN THE VOLTAGE CROSSES THRESHOLD) IS SET TO NAN.
        #     THE INTERPOLATED VOLTAGE CAN BE USED BUT THEN THE INTERPOLATED VOLTAGE MUST BE CALCULATED FOR THE TRUE VOLTAGE 
        #     TRACE AND POSSIBLY IN THE INTEGRATION.
        if fake:
            v_comp_of_th_at_each_spike_via_data=v_trace[all_spikeInd-1]-internal_sp_comp_array[all_spikeInd-1]-th_inf  #USE THIS FOR FAKE DATA
        else:
            v_comp_of_th_at_each_spike_via_data=v_trace[all_spikeInd]-internal_sp_comp_array[all_spikeInd]-th_inf  #USE THIS FOR REAL DATA (although probably not necessary)
        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        
        # For each ISI, calculate the difference between the voltage dependent component of the threshold
        # and the value that would be determined via a model that uses the actual voltage of neuron.  
        sq_err = []    #list to store squared error between the model and biological threshold
        for spike_number in range(1, len(all_spikeInd)):      #loop over all ISI's in data
            v_start_ind = all_spikeInd[spike_number-1]+int(spike_cut_length) #dont want to use voltage during a spike
            end_ind = all_spikeInd[spike_number]
            v_in_ISI=v_trace[v_start_ind:end_ind]
            #voltage component of threshold at the beginning and end of the ISI
            #With fake data if go from one before fake data to fake data this should be exact
            theta0=v_comp_of_th_at_each_spike_via_data[spike_number-1] #this assumes that the voltage component of the threshold does not change over the time period of the spike
            # not sure this makes sense any moretheta0=v_comp_of_th_at_each_spike_via_data[spike_number-1]+sp_comp_of_offset_at_spike_list[spike_number-1] #use this is you want to add the biological component back on
            theta1=v_comp_of_th_at_each_spike_via_data[spike_number]
            tvec=np.arange(len(v_in_ISI))*dt
    
            #analytical solution should be exact with fake data--small differences could be because of the differences in the voltage at 
            #spike indicies are off by one
            model=+theta0*np.exp(-b_voltage*dt*(end_ind-v_start_ind))+a_voltage*np.exp(-b_voltage*tvec[-1])*np.sum(dt*(v_in_ISI-El)*np.exp(b_voltage*tvec))
            err = (theta1-model)**2
            if ~np.isnan(err):
                sq_err.append(err)
        
        total_err+=np.sum(sq_err)
    return total_err

def fit_avoltage_bvoltage_th(x, v_trace_list, El_list, spike_cut_length, all_spikeInd_list, dt, a_spike, 
                     b_spike, fake=False):
    '''This is a version of fit_avoltage_bvoltage_th_debug that does not require the th_trace, 
    v_component_of_thresh_trace, and spike_component_of_thresh_trace needed for debugging. A
    test should be run to make sure the same output comes out from this and the debug function
    
    This function returns the squared error for the difference between the 'known' voltage
    component of the threshold obtained from the biological neuron and the voltage component 
    of the threshold of the model obtained with the input parameters (so that the minimum can be 
    searched for via fmin). The overall threshold is the sum of threshold infinity the spike component
    of the threshold and the voltage component of the threshold.  Therefore threshold infinity and 
    the spike component of the threshold must be subtracted from the threshold of the neuron in order
    to isolate the voltage component of the threshold.  In the evaluation of the model the actual
    voltage of the neuron is used so that any errors in the other components of the model will not 
    influence the fits here (for example, if a afterspike current was estimated incorrectly)
  
    Notes:
    * The spike component of the threshold is subtracted from the 
        voltage which means that the voltage component of the threshold should only be added to rules.
    * b_spike was fit using a negative value in the function therefore the negative is placed in the 
        equation. 
    * values in this function are in 'real' voltage as opposed to voltage
        relative to resting potential. 
    * current injection during the spike is not taken into account.  This seems reasonable as the 
        ion channels are open during this time and injected current may not greatly influence the neuron.
    
    x: numpy array
        x[0]=a_voltage input, x[1] is b_voltage_input, x[2] is th_inf
    v_trace_list: list of numpy arrays
        voltage traces (v_trace, El, and th_inf must be in the same frame of reference)
    El_list: list of floats
        reversal potential (v_trace, El, and th_inf must be in the same frame of reference)
    spike_cut_length: int
        number of indicies removed after initiation of a spike
    all_spikeInd_list: list of numpy arrays
        indicies of spike trains 
    dt: float 
        size of time step (SI units)
    a_spike: float
        amplitude of spike component of threshold.
    b_spike: float
        decay constant in spike component of the threshold
    fake: Boolean
        if True makes uses the voltage value of spike step-1 because there is not a voltage value at the spike
        step because it is set to nan in the simulator.
    '''
    a_voltage=x[0]
    b_voltage=x[1]
    th_inf=x[2]

    total_err=0
    for v_trace, El, all_spikeInd in zip(v_trace_list, El_list, all_spikeInd_list):
        # Calculate values along the whole trace and then take the values at the spike ind
        internal_sp_comp_array=np.zeros(all_spikeInd[0]+spike_cut_length) 
        left_over=0
        #vector of spike component of of the threshold from each spike and previous spikes; note at first spike there is no spike component of threshold so initialized at zero   
        #Note that care has to be taken here to get make sure the right amount of decay is left over
        for spike_number in range(1,len(all_spikeInd)):      #skipping first spike since no residual spike component of threshold
            integration_length=all_spikeInd[spike_number]-all_spikeInd[spike_number-1]+1 #this is the amount of time that needs to be integrated over note that it is one longer than the interval because of the last value is added to the aspike for the next ISI
            local_spike_comp_of_threshold=spike_component_of_threshold_exact(a_spike+left_over, b_spike, np.arange(integration_length)*dt)
            internal_sp_comp_array=np.append(internal_sp_comp_array, local_spike_comp_of_threshold[:-1])
            left_over=local_spike_comp_of_threshold[-1]
        
        # Compute voltage component of threshold at biological spike (subtract th_inf and spike component of threshold
        # from biological voltage values at spike initiation)
        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        #     NOTE THAT THERE IS AN ISSUE HERE USING FAKE DATA.  THE -1 IS HERE BECAUSE THE NEURON CROSSES THRESHOLD SOMETIME BETWEEN TWO INDICIES.
        #     FOR THE FAKE DATA THE TIME OF THE SPIKE (THE POINT FOLLOWING WHEN THE VOLTAGE CROSSES THRESHOLD) IS SET TO NAN.
        #     THE INTERPOLATED VOLTAGE CAN BE USED BUT THEN THE INTERPOLATED VOLTAGE MUST BE CALCULATED FOR THE TRUE VOLTAGE 
        #     TRACE AND POSSIBLY IN THE INTEGRATION.
        if fake:
            v_comp_of_th_at_each_spike_via_data=v_trace[all_spikeInd-1]-internal_sp_comp_array[all_spikeInd-1]-th_inf  #USE THIS FOR FAKE DATA
        else:
            v_comp_of_th_at_each_spike_via_data=v_trace[all_spikeInd]-internal_sp_comp_array[all_spikeInd]-th_inf  #USE THIS FOR REAL DATA (although probably not necessary)
        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        
        # For each ISI, calculate the difference between the voltage dependent component of the threshold
        # and the value that would be determined via a model that uses the actual voltage of neuron.  
        sq_err = []    #list to store squared error between the model and biological threshold
        for spike_number in range(1, len(all_spikeInd)):      #loop over all ISI's in data
            v_start_ind = all_spikeInd[spike_number-1]+int(spike_cut_length) #dont want to use voltage during a spike
            end_ind = all_spikeInd[spike_number]
            v_in_ISI=v_trace[v_start_ind:end_ind]
            #voltage component of threshold at the beginning and end of the ISI
            #With fake data if go from one before fake data to fake data this should be exact
            theta0=v_comp_of_th_at_each_spike_via_data[spike_number-1] #this assumes that the voltage component of the threshold does not change over the time period of the spike
            # not sure this makes sense any moretheta0=v_comp_of_th_at_each_spike_via_data[spike_number-1]+sp_comp_of_offset_at_spike_list[spike_number-1] #use this is you want to add the biological component back on
            theta1=v_comp_of_th_at_each_spike_via_data[spike_number]
            tvec=np.arange(len(v_in_ISI))*dt
    
            #analytical solution should be exact with fake data--small differences could be because of the differences in the voltage at 
            #spike indicies are off by one
            model=+theta0*np.exp(-b_voltage*dt*(end_ind-v_start_ind))+a_voltage*np.exp(-b_voltage*tvec[-1])*np.sum(dt*(v_in_ISI-El)*np.exp(b_voltage*tvec))
            err = (theta1-model)**2
            if ~np.isnan(err):
                sq_err.append(err)
        
        total_err+=np.sum(sq_err)
    return total_err


#TODO: depricate confirmed use of fit_avoltage_bvoltage_th
#def err_fix_th(x, v_trace, El, spike_cut_length, all_spikeInd, th_inf, dt, a_spike, b_spike):
#    '''This function returns the squared error for the difference between the 'known' voltage 
#    component of the threshold obtained from the biological neuron and the voltage component 
#    of the threshold of the model obtained with the input parameters (so that the minimum can be 
#    searched for via fmin). The overall threshold is the sum of threshold infinity the spike component
#    of the threshold and the voltage component of the threshold.  Therefore threshold infinity and 
#    the spike component of the threshold must be subtracted from the threshold of the neuron in order
#    to isolate the voltage component of the threshold.  In the evaluation of the model the actual
#    voltage of the neuron is used so that any errors in the other components of the model will not 
#    influence the fits here (for example if a afterspike current was estimated incorrectly)
#  
#    Notes:
#    * The spike component of the threshold is subtracted from the 
#        voltage which means that the voltage component of the threshold should only be added to rules.
#    * b_spike was fit using a negative value in the function therefore the negative is placed in the 
#        equation. 
#    * values in this function are in 'real' voltage as opposed to voltage
#        relative to resting potential. 
#    * current injection during the spike is not taken into account.  This seems reasonable as the 
#        ion channels are open during this time and injected current may not greatly influence the neuron.
#    
#    x: numpy array
#        x[0]=a_voltage input, x[1] is b_voltage_input
#    voltage: numpy array
#        voltage trace (voltage, El, and th_inf must be in the same frame of reference)
#    El: float
#        reversal potential (voltage, El, and th_inf must be in the same frame of reference)
#    spike_cut_length: int
#        number of indicies removed after initiation of a spike
#    all_spikeInd: numpy array
#        indicies of spike train 
#    th_inf: float
#        threshold infinity (voltage, El, and th_inf must be in the same frame of reference)
#    dt: float 
#        size of time step (SI units)
#    a_spike: float
#        amplitude of spike component of threshold.
#    b_spike: float
#        decay constant in spike component of the threshold
#    '''
#    a_voltage=x[0]
#    b_voltage=x[1]
#    # effect of the spike component of the threshold from each spike and previous spikes
#    sp_comp_of_offset_sum_vector=[0] #vector of spike component of of the threshold from each spike and previous spikes; note at first spike there is no spike component of threshold so initialized at zero   
#    for spike_number in range(1,len(all_spikeInd)):      #skipping first spike since no residual spike component of threshold
#        t=(all_spikeInd[spike_number]-all_spikeInd[spike_number-1]-spike_cut_length)*dt #spike ISI
#        sp_comp_of_th_offset_local = spike_component_of_threshold_exact(a_spike, b_spike, t) #spike component of threshold at each ISI for each individual spike
#        #I THINK THE LINE BELOW MIGHT JUST BE WRONG BECAUSE THE OLD OFF SET WOULD DECAY AND I DONT THINK IT IS HERE:THIS IS WHAT IS BEING USED
##        sp_comp_of_offset_sum_vector.append(sp_comp_of_offset_sum_vector[-1] + sp_comp_of_th_offset_local)  #keeping track of residual spike component of threshold at each spike
#        left_over_decay=spike_component_of_threshold_exact(sp_comp_of_offset_sum_vector[-1], b_spike, t)
#        sp_comp_of_offset_sum_vector.append(left_over_decay + sp_comp_of_th_offset_local)  #keeping track of spike component of threshold with residuals at each spike
#
#
#    # Compute v_trace component of threshold at biological spike (subtract th_inf and spike component of threshold
#    # from biological v_trace values at spike initiation)
#    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#    # NOTE THAT THERE IS AN ISSUE HERE USING FAKE DATA.  THE -1 IS HERE BECAUSE THE NEURON CROSSES THRESHOLD SOMETIME BETWEEN TWO INDICIES.
#    # FOR THE FAKE DATA THE TIME OF THE SPIKE (THE POINT FOLLOWING WHEN THE VOLTAGE CROSSES THRESHOLD) IS SET TO NAN.
#    # THE INTERPOLATED VOLTAGE CAN BE USED BUT THEN THE INTERPOLATED VOLTAGE MUST BE CALCULATED FOR THE TRUE VOLTAGE 
#    # TRACE AND POSSIBLY IN THE INTEGRATION.
#    v_comp_of_th_at_each_spike_via_data=v_trace[all_spikeInd-1]-np.array(sp_comp_of_offset_sum_vector)-th_inf  #USE THIS FOR FAKE DATA
##    v_comp_of_th_at_each_spike_via_data=v_trace[all_spikeInd]-np.array(sp_comp_of_offset_sum_vector)-th_inf  #THIS IS PROBABLY APPROPRIATE FOR REAL DATA
#    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#    
#    # For each ISI, calculate the difference between the v_trace dependent component of the threshold
#    # and the value that would be determined via a model that uses the actual v_trace of neuron.  
#    sq_err = []    #list to store squared error between the model and biological threshold
#    for spike_number in range(1, len(all_spikeInd)):      #loop over all ISI's in data
#        v_start_ind = all_spikeInd[spike_number-1]+int(spike_cut_length) #dont want to use voltage during a spike
#        end_ind = all_spikeInd[spike_number]
#        v_in_ISI=v_trace[v_start_ind:end_ind]
#        #v_trace component of threshold at the beginning and end of the ISI
#    
#        theta0=v_comp_of_th_at_each_spike_via_data[spike_number-1] #this assumes that the voltage component of the threshold does not change over the time period of the spike
#        # not sure this makes sense any moretheta0=v_comp_of_th_at_each_spike_via_data[spike_number-1]+sp_comp_of_offset_sum_vector[spike_number-1] #use this is you want to add the biological component back on
#        theta1=v_comp_of_th_at_each_spike_via_data[spike_number]
#        tvec=np.arange(len(v_in_ISI))*dt
#
#        #need to prove this--NOTE THAT THIS SHOULD BE EXACT DURING FAKE DATA--DIFFERENCES COULD BE DUE TO GETTING VALUES AT DIFFERENT INDICIES
#        model=+theta0*np.exp(-b_voltage*dt*(end_ind-v_start_ind))+a_voltage*np.exp(-b_voltage*tvec[-1])*np.sum(dt*(v_in_ISI-El)*np.exp(b_voltage*tvec))
#        err = (theta1-model)**2
#        if ~np.isnan(err):
#            sq_err.append(err)
#             
#    return np.sum(sq_err)

# TODO: can depricate, using sdk now
#def find_multiblip_spikes(multi_SS_i, multi_SS_v, dt):
#    '''artifacts caused by turning stimulus on and off created artifacts that 
#    created problems for the standard spike detection algorithm.  Several alterations 
#    were made so that the algorithm would detect spikes appropriately. Please see multiblip
#    spike cutting documentation for more information on how the code differs from the SDK 
#    version and what was needed to solve specific issues.
#    input:
#    multi_SS_i: list of arrays
#        each array corresponds to the current stimulation of one triblip stimulus
#    multi_SS_v: list of arrays
#        each array corresponds to the voltage trace of triblip simulus
#    dt: float 
#    '''
#    artifact_ave_window_time_s=0.0003 #
#    window_indicies_to_ave_len=int(artifact_ave_window_time_s/dt)
#    out_spk_idxs_list=[]
#    for current, voltage in zip(multi_SS_i, multi_SS_v):
#        #--Find the beginning and end of stimuli so that we can average the voltage traces at that time
#        up_blip_index=np.where(np.diff(np.greater_equal(current, 1e-10).astype(int)) == 1)[0]#Note 1e-10 is larger than test pulse so it will not pick up test pulse
#        down_blip_index=np.where(np.diff(np.less_equal(current, 1e-10).astype(int)) == 1)[0]
#        potential_artifact_indexes=np.sort(np.append(up_blip_index, down_blip_index))
#        artifact_removed_voltage=copy.deepcopy(voltage)
#        window_boarders_index=[]
#        #--remove artifacts 
#        for index in potential_artifact_indexes:
#            smooth_window=range(index,index+window_indicies_to_ave_len+1)
#            
#            # interpolate in the smoothing window
#            blah=interp1d([smooth_window[0], smooth_window[-1]], [artifact_removed_voltage[smooth_window[0]], artifact_removed_voltage[smooth_window[-1]]])
#            artifact_removed_voltage[smooth_window]=blah(smooth_window)
#            
#            #windows boarders are just for plotting
##            window_boarders_index.append(smooth_window[0])
##            window_boarders_index.append(smooth_window[-1])
##        plt.figure()
##        plt.plot(voltage, 'b', lw=4)
##        plt.plot(artifact_removed_voltage, 'r', lw=2)
##        plt.plot(window_boarders_index, artifact_removed_voltage[window_boarders_index], '|g', ms=10)
##        plt.xlim([40400, 41000])
##        plt.show()
##        
##        t = np.arange(0, len()) * dt
#
#        # keeping the smooth_v convention of the SDK find spike code.  However in the SDK code 
#        # this is used to name data potentially smoothed by a bessel filter
#        smooth_v = artifact_removed_voltage
#        dv = np.diff(smooth_v)    
#        dvdt = dv / dt
#        dvv = np.diff(dvdt)
#                
#        v=smooth_v[:-1]  #truncating the end of v so it has the same dimensions as dvdt for time plotting
#
#        spikes = []
#        out_spk_idxs = []
#        
#        peaks=get_peaks(v) # find potential spikes by finding peak over zero mv
#      
#        # Etay defines spike as time of threshold crossing.  Threshold is defined as the time at which dvdt is some percent of maximum threshold.
#        # TODO: figure out how maximum threshold is defined in original code so I can say why I don't use it
#        for spk_n, peak_idx in enumerate(peaks):
#            #---------find spike peak----------------------------
#            spk = {}
#    
#            spk["peak_idx"] = peak_idx 
#            upstroke_idx = np.argmax(dvdt[peak_idx-int(.001/dt):peak_idx]) + peak_idx-int(.001/dt)  
#            spk["upstroke"] = dvdt[upstroke_idx]
#            spk["upstroke_idx"] = upstroke_idx
#            spk["upstroke_v"] = v[upstroke_idx]
#                
#            # Define threshold where dvdt = 5% * max upstroke
#            dvdt_thr_target = THRESH_PCT_MULTIBLIP * spk["upstroke"]
#            #print 'spk[upstroke]', spk["upstroke"], 'dvdt_thr_target', dvdt_thr_target
#            prev_idx = peak_idx-int(.0035/dt)
#            #check to make sure prev_idx is not before or in a window where the stimulus blip comes on because 
#            #it will errorniously trip the threshold dvdt
#            for index in up_blip_index:
#                if prev_idx<=index+int(.0005/dt) and prev_idx>= index-int(.0035/dt):
#                    prev_idx=index+int(.0005/dt)
#                    
#            mean_dvv= [np.mean(dvv[pv-2:pv+3]) for pv in range(prev_idx,upstroke_idx)] #makes sure dv2/dt2 isnt spuriously going down by averaging 5 points
#            find_thresh_idxs = np.where(np.logical_and(dvdt[prev_idx:upstroke_idx] >= dvdt_thr_target,  np.greater(mean_dvv,0)))[0]  
#               
#            if len(find_thresh_idxs) < 1: # Can't find a good threshold value - probably a bad simulation case
#                # Fall back to the upstroke value
#                threshold_idx = upstroke_idx
#            else:
#                threshold_idx = find_thresh_idxs[0] + prev_idx
#                    
#            spk["threshold_idx"] = threshold_idx
#            spk["threshold_v"] = v[threshold_idx]
#                
#            # Check for things that are probably not spikes:
#  
#            # if the "spike" is less than 2 mV from threshold to peak, don't count it
#            if v[peak_idx] - v[threshold_idx] < 0.002:  
#                print "\tnot counting spike is closer to peak than 2 mV"
#                continue
#    
#            #NOTE: because threshold doesnt decay to zero in the multiblip this doesnt get rid of the situation that usually is only in the first spike of a stimulus
#            # if the spike is less the -30mV, don't count it
#            if v[peak_idx] < -0.04:
#                print "\tnot counting spike: peak is too small"
#                continue
#    
#            spikes.append(spk)
#    
#    #----figure out if I should still find a global threshold and then do it all again
#    #    # find global threshold which is an average of the individual thresholds
#    #    if len(spikes) > 0:
#    #        dvdt_thr_target = np.array([spk["upstroke"] for spk in spikes]).mean() * THRESH_PCT_MULTIBLIP
#    #    else: # if there weren't any spikes, move along
#    #        return np.array([])
#        
#            out_spk_idxs.append(spk["threshold_idx"])
#        
#        out_spk_idxs_list.append(np.array(out_spk_idxs))
#        
##        time_vector=np.arange(len(v))*dt
##        plt.figure()
##    #    plt.subplot(3,1,1)
##    #    plt.plot(time_vector, ddv)
##    #    plt.plot(time_vector[out_spk_idxs], ddv[out_spk_idxs], '.r', ms=16)
##    #    plt.xlim([40300, 42000])
##    #    plt.ylabel('ddv')
##        plt.subplot(2,1,1)
##        plt.plot(time_vector, dvdt)
##        plt.plot(time_vector[out_spk_idxs], dvdt[out_spk_idxs], '.r', ms=16)
##        plt.plot(time_vector[potential_artifact_indexes], dvdt[potential_artifact_indexes], 'b|', ms=24, lw=4)
##        plt.xlim([40300*dt, 42000*dt])
##        plt.ylabel('dvdt')
##        plt.subplot(2,1,2)
##        plt.plot(time_vector, v)
##        plt.plot(time_vector[out_spk_idxs], v[out_spk_idxs], 'r.', ms=16, label='threshold')
##        plt.xlim([40300*dt, 42000*dt])
##        plt.ylabel('voltage (V)')
##        plt.plot(time_vector[peaks], v[peaks], '.g', ms=16, label='peaks')
##        plt.plot(time_vector[[spikes[ii]['upstroke_idx'] for ii in range(len(spikes))]], [spikes[ii]['upstroke_v'] for ii in range(len(spikes))], '.c', ms=16, label = 'max upstroke')
##        plt.plot(time_vector[potential_artifact_indexes], v[potential_artifact_indexes], 'b|', ms=24, lw=4)
##        plt.legend()
##        plt.show()
#        
#    return out_spk_idxs_list

def get_peaks(voltage, aboveValue=0):
    '''This function was written by Corinne Teeter and calculates the action potential peaks of a voltage equation"
    inputs
        voltage: numpy array of voltages
        aboveValue: scalar voltage value over which voltage is considered a spike.
    outputs:    
        peakInd: array of indicies of peaks'''
    VshiftR=np.concatenate(([0], voltage[0:voltage.size-2]))
    VshiftL=voltage[1:voltage.size]
    IndShiftR=np.where(voltage[0:voltage.size-1]>VshiftR)
    IndShiftL=np.where(voltage[0:voltage.size-1]>VshiftL)
    greatThanThresh=np.where(voltage>aboveValue) #finds indicies greater than the value provided
    peakInd=np.intersect1d(np.intersect1d(IndShiftL[0], IndShiftR[0]), greatThanThresh[0]) #find the indicies of the peak
    return peakInd

def exp_force_c((t, const), a1, k1):
    return a1*(np.exp(k1*t))+const

def exp_fit_c(t, a1, k1, const):
    return a1*(np.exp(k1*t))+const




from __future__ import division
import logging

import numpy as np
from six import iteritems

from .interval_unionizer import IntervalUnionizer
from .tissuecyte_unionize_record import TissuecyteInjectionUnionize, \
    TissuecyteProjectionUnionize

  
class TissuecyteUnionizer(IntervalUnionizer):
    '''A specialization of the IntervalUnionizer set up for unionizing 
    Tissuecyte-derived projection data.
    '''
    

    @classmethod
    def record_cb(cls):
        return {'injection': TissuecyteInjectionUnionize(), 
                'projection': TissuecyteProjectionUnionize()}
    
    
    def extract_data(self, data_arrays, low, high):
        '''As parent
        '''
    
        unionize = self.__class__.record_cb()

        unionize['injection'].calculate(low, high, data_arrays)
        unionize['projection'].calculate(low, high, data_arrays, unionize['injection'])
        
        return unionize 
        
    
    @classmethod                
    def propagate_record(cls, child_record, ancestor_record, copy_all=False):
        '''As parent
        '''

        for k, v in child_record.iteritems():
            v.propagate(ancestor_record[k], copy_all)
        
        return ancestor_record
        

    def postprocess_unionizes(self, raw_unionizes, image_series_id, 
                              output_spacing_iso, volume_scale, target_shape, sort):
        '''As parent
        
        New Parameters
        --------------
        output_spacing_iso : numeric
            Isometric spacing of reference space in microns
        volume_scale : numeric
            Scale factor mapping pixels to microns^3
        target_shape : array-like of numeric
            Shape of reference space
        
        '''

        unionizes = []
        total_injection_volume = 0
        
        logging.info('getting formatted unionize output')
        for sid, un in raw_unionizes.iteritems():
            
            if sid < 0:
                hemisphere = 1
            else:
                hemisphere = 2
                
            current = []
            for ij, item in iteritems(un):    
            
                v = item.output(output_spacing_iso, volume_scale, target_shape, sort)
                injection = True if ij == 'injection' else False
                
                if injection and hemisphere != 3:
                    total_injection_volume += v['direct_projection_volume']
                    
                del v['direct_projection_volume']
                del v['direct_sum_projection_pixels']
            
                v.update({'is_injection': injection, 
                          'hemisphere_id': hemisphere, 
                          'structure_id': abs(sid), 
                          'image_series_id': image_series_id})
                          
                current.append(v)
                
            unionizes.extend(current)
        
        if total_injection_volume > 0:
            logging.info('computing normalized projection volume')
            for un in unionizes:
                un['normalized_projection_volume'] = un['projection_volume'] / total_injection_volume
        else:
            logging.warning('no injection found!')
            for un in unionizes:
                un['normalized_projection_volume'] = 0
            
        return filter(lambda x: x['sum_pixels'] > 0, unionizes)

    
    

    



import subprocess as sp
import numpy as np
import logging
import sys, os
from collections import deque
import scipy.misc
import traceback
import signal

class FrameInputStream( object ):
    def __init__(self, movie_path, num_frames=None, block_size=1, cache_frames=False, process_frame_cb=None):
        self.movie_path = movie_path
        self.num_frames = num_frames
        self.block_size = block_size
        self.cache_frames = cache_frames
        self.process_frame_cb = process_frame_cb if process_frame_cb else lambda f: f[:,:,0].copy()

        self.frames_read = 0
        self.frame_cache = []

    def open(self):
        self.frames_read = 0

    def close(self):
        logging.debug("Read total frames %d", self.frames_read)

        if self.num_frames is not None and self.frames_read != self.num_frames:
            raise IOError("read incorrect number of frames: %d vs %d", self.frames_read, self.num_frames)

    def _error(self):
        pass

    def _process_frame(self, frame):
        return self.process_frame_cb(frame)
    
    def _read_iter(self):
        pass

    def __enter__(self):
        return self

    def __iter__(self):
        # if we're caching frames and the cache exists, return it
        if self.cache_frames and self.frame_cache:
            n = self.num_frames if self.num_frames is not None else len(self.frame_cache)
            for i in range(n):
                yield self.frame_cache[i]
        else:
            self.open()

            self.frame_cache = []

            for frame in self._read_iter():
                self.frame_cache.append(self._process_frame(frame))
                self.frames_read += 1
                
                if (self.frames_read % 100) == 0:
                    logging.debug("Read frames %d", self.frames_read)

                if self.block_size is None:
                    continue
                if self.block_size == 1:
                    yield self.frame_cache[-1]
                elif (self.frames_read % self.block_size) == 0:
                    for i in range(-self.block_size,0):
                        yield self.frame_cache[i]

                    if not self.cache_frames:
                        self.frame_cache = []

            self.close()

            for frame in self.frame_cache:
                yield frame

            if not self.cache_frames:
                self.frame_cache = []


    def __exit__(self, exc_type, exc_value, tb):
        if exc_value:
            traceback.print_tb(tb)
            self._error()
            raise exc_value

    def create_images(self, output_directory, image_type):
        for i, frame in enumerate(self):
            file_name = os.path.join(output_directory, "input_frame-%06d." % i  + image_type)
            scipy.misc.imsave(file_name, frame)
        

class FfmpegInputStream( FrameInputStream ):
    def __init__(self, movie_path, frame_shape, ffmpeg_bin='ffmpeg', num_frames=None, block_size=1, cache_frames=False, process_frame_cb=None):
        super(FfmpegInputStream, self).__init__(movie_path=movie_path, num_frames=num_frames, block_size=block_size, cache_frames=cache_frames, process_frame_cb=process_frame_cb)

        self.ffmpeg_bin = ffmpeg_bin
        self.frame_shape = frame_shape

        self.pipe = None
        
    def open(self):
        super(FfmpegInputStream, self).open()

        if self.pipe:
            raise IOError("pipe is open already")

        command = [ self.ffmpeg_bin,
                    '-i', self.movie_path,
                    '-f', 'image2pipe',
                    '-pix_fmt', 'rgb24',
                    '-vcodec', 'rawvideo']

        if self.num_frames is not None:
            command += ['-vframes', str(self.num_frames)]

        command += ['-']

        frame_size = np.prod(self.frame_shape)
        self.pipe = sp.Popen(command, stdout=sp.PIPE, bufsize=0)
        logging.debug("opened pipe")

    def close(self):
        if self.pipe is None:
            raise IOError("pipe is not open")

        if self.pipe.poll() is None:
            logging.debug("pipe is still open. terminating.")
            self.pipe.terminate()

        super(FfmpegInputStream, self).close()


        rc = self.pipe.wait()
        logging.debug("closed input pipe")
        
        if rc:
            raise Exception("input pipe returned with error code %d" % rc)        

        self.pipe = None

    def _process_frame(self, frame):
        frame = np.fromstring(frame, dtype=np.uint8)
        frame.resize(self.frame_shape)
        return self.process_frame_cb(frame)

    def _read_iter(self):
        if self.pipe is None:
            raise IOError("pipe is not open")

        frame_size = np.prod(self.frame_shape)

        while self.pipe.poll() is None or self.pipe.stdout:
            self.pipe.stdout.flush()
            input_frame = self.pipe.stdout.read(frame_size)

            bytes_read = len(input_frame)

            if bytes_read == 0:
                break

            if bytes_read != frame_size:
                raise IOError("pipe read wrong number of bytes (%d vs %d)" % (frame_size, bytes_read))

            yield input_frame

    def _error(self):
        if self.pipe:
            self.pipe.kill()
            self.pipe = None

    def create_images(self, output_directory, image_type):
        cmd = self.ffmpeg_bin + ' -i ' + self.movie_path + ' ' + output_directory + '/input_frame-%06d.' + image_type

        logging.debug("Calling ffmpeg with the command:")
        logging.debug("\t"+cmd)
        retcode = sp.call(cmd, shell=True)
        if retcode != 0:
            logging.debug(retcode)
            raise Exception('Something went wrong with image creation')



class CvInputStream( object):
    def __init__(self, movie_path, num_frames=None, block_size=1, cache_frames=False):
        super(FfmpegInputStream, self).__init__(movie_path=movie_path, num_frames=num_frames, block_size=block_size, cache_frames=cache_frames)
        self.cap = None
        
    def open(self):
        super(FfmpegInputStream, self).open()

        if self.cap:
            raise IOError("capture is open already")

        self.frames_read = 0

        import cv2
        self.cap = cv2.VideoCapture(self.movie_path)
        logging.debug("opened capture")

    def close(self):
        if self.cap is None:
            return

        self.cap.release()
        self.cap = None

        super(FfmpegInputStream, self).close()

    def _read_iter(self):
        if self.cap is None:
            raise IOError("capture is not open")

        while self.cap.isOpened():
            ret, frame = self.cap.read()
            yield frame

            if self.frames_read == self.num_frames:
                break

    def _error(self):
        self.cap.release()
        self.cap = None

class FrameOutputStream( object ):
    def __init__(self, block_size=1):
        self.frames_processed = 0
        self.block_frames = []
        self.block_size = block_size

    def open(self, movie_path):
        self.frames_processed = 0
        self.block_frames = []
        self.movie_path = movie_path

    def _write_frames(self, frames):
        raise NotImplementedError()

    def write(self, frame):
        self.block_frames.append(frame)

        if len(self.block_frames) == self.block_size:
            self._write_frames(self.block_frames)
            self.frames_processed += len(self.block_frames)
            self.block_frames = []

    def close(self):
        if self.block_frames:
            self._write_frames(self.block_frames)
            self.frames_processed += len(self.block_frames)
            self.block_frames = []

        logging.debug("wrote %d frames", self.frames_processed)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_value:
            raise exc_value
        self.close()

class ImageOutputStream( FrameOutputStream ):
    def _write_frames(frames):
        for i, frame in enumerate(frames):
            file_name = self.movie_path % i
            scipy.misc.imsave(file_name, frame)
        

class FfmpegOutputStream( FrameOutputStream ):
    def __init__(self, frame_shape, ffmpeg_bin='ffmpeg', block_size=1):
        super(FfmpegOutputStream, self).__init__(block_size)

        self.ffmpeg_bin = ffmpeg_bin
        self.frame_shape = frame_shape
        self.pipe = None
        self.stopped = False

    def open(self, movie_path):
        super(FfmpegOutputStream, self).open(movie_path)

        if self.pipe:
            logging.warning("pipe is already open!")
            return

        command = [ self.ffmpeg_bin,
                    '-y',
                    '-f', 'rawvideo',
                    '-vcodec', 'rawvideo',
                    '-s', '%dx%d' % (self.frame_shape[1], self.frame_shape[0]),
                    '-pix_fmt', 'rgb24',
                    '-r', '30',
                    '-i', '-',
                    '-an',
                    '-vcodec', 'libx264',
                    self.movie_path]

        self.pipe = sp.Popen(command, stdin=sp.PIPE)
        os.kill(self.pipe.pid, signal.SIGSTOP)
        self.stopped = True
        logging.debug("opened output pipe")


    def _write_frames(self, frames):
        if self.pipe is None:
            self.open(self.movie_path)
        if self.stopped:
            os.kill(self.pipe.pid, signal.SIGCONT)
            self.stopped = False

        for frame in frames:
            sys.stdout.flush()
            self.pipe.stdin.write( frame.tostring() )
        
    def close(self):
        super(FfmpegOutputStream, self).close()
        if self.pipe is None:
            raise IOError("pipe is closed")

        self.pipe.stdin.close()
        rc = self.pipe.wait()

        if rc:
            raise Exception("output pipe returned with error code %d" % rc)

        logging.debug("closed output pipe")
        self.pipe = None

    def __exit__(self, exc_type, exc_value, tb):
        if exc_value:
            self.pipe.kill()
            raise exc_value
        self.close()
        




import numpy as np
import logging

class FitEllipse (object):

    def __init__(self,min_points,max_iter,threshold,num_close):

        # points = np.array(candidate_points)
        # y,x = points.T

        C = np.zeros([6,6])
        C[0,2]= 2.0
        C[2,0]= 2.0
        C[1,1]= -1.0

        #self.x = x
        #self.y = y
        self.C = C

        self.min_points = min_points
        self.max_iter = max_iter
        self.threshold = threshold
        self.num_close = num_close

        self.best_params = None
        self.best_params_set = False
        self.besterror = np.inf

    def ransac_fit(self,candidate_points):

        #points = np.array(candidate_points)

        for i in range(self.max_iter):

            inlier_points, outlier_points = self.choose_inliers(candidate_points)
            params, error = self.fit_ellipse(inlier_points)

            if len(outlier_points)>0:
                cost = self.outlier_cost(outlier_points,params)
                also_in = 0
                for j,c in enumerate(cost):
                    point = outlier_points[j]
                    if cost[j]<self.threshold:
                        inlier_points += [point]
                        also_in += 1

                if also_in > self.num_close:
                    params, error = self.fit_ellipse(inlier_points)
                    if (error < self.besterror):
                        self.best_params = params
                        self.best_params_set = True
                        self.besterror = error

        if self.best_params_set:
            return ellipse_center(self.best_params), ellipse_angle_of_rotation(self.best_params)*180./np.pi, ellipse_axis_length(self.best_params)
        else:
            return None

    def choose_inliers(self, candidate_points):

        #cannot take a larger sample than population
        if(len(candidate_points) > self.min_points):
            inlier_index = np.random.choice(np.arange(len(candidate_points)),self.min_points,replace=False)
        else:
            #TODO check this
            inlier_index = np.arange(self.min_points)

        inlier_points = []
        outlier_points = []

        for i in range(len(candidate_points)):
            if i in inlier_index:
                inlier_points += [candidate_points[i]]
            else:
                outlier_points += [candidate_points[i]]

        return inlier_points, outlier_points

    def outlier_cost(self,outlier_points,params):

        y,x = np.array(outlier_points).T

        D = np.vstack([x*x, x*y, y*y, x, y, np.ones(len(y))])
        #S = np.dot(D, D.T)

        cost = (np.dot(params,D))**2

        return cost

    def fit_ellipse(self,inlier_points):
        try:
            inlier_points = np.array(inlier_points)
            points = np.array(inlier_points)
            y,x = points.T

            D = np.vstack([x*x, x*y, y*y, x, y, np.ones(len(y))])
            S = np.dot(D, D.T)

            M = np.dot(np.linalg.inv(S),self.C)
            U,s,V=np.linalg.svd(M)

            params = U.T[0]
            error = np.dot(params, np.dot(S,params))/len(inlier_points)
        except:
            #TODO - check if this is correct
            params = None      #WBW error handling
            error = 0.00000001 #WBW error handling

        return params, error


def ellipse_center(a):
    b,c,d,f,g,a = a[1]/2, a[2], a[3]/2, a[4]/2, a[5], a[0]
    num = b*b-a*c
    x0=(c*d-b*f)/num
    y0=(a*f-b*d)/num
    return np.array([x0,y0])

def ellipse_angle_of_rotation( a ):
    b,c,d,f,g,a = a[1]/2, a[2], a[3]/2, a[4]/2, a[5], a[0]
    return 0.5*np.arctan(2*b/(a-c))

def ellipse_angle_of_rotation2( a ):
    b,c,d,f,g,a = a[1]/2, a[2], a[3]/2, a[4]/2, a[5], a[0]
    if b == 0:
        if a > c:
            return 0
        else:
            return np.pi/2
    else:
        if a > c:
            return np.arctan(2*b/(a-c))/2
        else:
            return np.pi/2 + np.arctan(2*b/(a-c))/2


def ellipse_axis_length( a ):
    b,c,d,f,g,a = a[1]/2, a[2], a[3]/2, a[4]/2, a[5], a[0]
    up = 2*(a*f*f+c*d*d+g*b*b-2*b*d*f-a*c*g)
    down1=(b*b-a*c)*( (c-a)*np.sqrt(1+4*b*b/((a-c)*(a-c)))-(c+a))
    down2=(b*b-a*c)*( (a-c)*np.sqrt(1+4*b*b/((a-c)*(a-c)))-(c+a))

    #TODO check this - cannot divide by 0 so just use a small number instead
    if(down1 == 0):
        down1 = .0000000001

    if(down2 == 0):
        down2 = .0000000001

    res1=np.sqrt(up/down1)
    res2=np.sqrt(up/down2)
    return np.array([res1, res2])

def fit_ellipse(candidate_points):

    # method from http://nicky.vanforeest.com/misc/fitEllipse/fitEllipse.html
    points = np.array(candidate_points)
    y,x = points.T
    D = np.vstack([x*x, x*y, y*y, x, y, np.ones(len(y))])
    S = np.dot(D, D.T)

    C = np.zeros([6,6])
    C[0,2]= 2.0
    C[2,0]= 2.0
    C[1,1]= -1.0

    M = np.dot(np.linalg.inv(S),C)

    U,s,V=np.linalg.svd(M)

    params = U.T[0]

    return ellipse_center(params), ellipse_angle_of_rotation(params)*180./np.pi, ellipse_axis_length(params)

def rotate_vector(y,x,theta):

    xp = x*np.cos(theta) - y*np.sin(theta)
    yp = x*np.sin(theta) + y*np.cos(theta)

    return yp,xp

def test_fit():

    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt

    x = np.linspace(-3.0,3.0,1000)
    yp = np.sqrt(4.0 - (4.0/9.0)*(x**2))
    ym = -yp

    yp += 0.1*np.random.normal(size=len(yp))
    ym += 0.1*np.random.normal(size=len(yp))

    yp, x1 = rotate_vector(yp,x,np.pi/8)
    ym, x2 = rotate_vector(ym,x,np.pi/8)

    y_outlier = np.random.random(size=100)*4.0 - 2.0
    x_outlier = np.random.random(size=100)*6.0 - 3.0


    outlier_points = np.vstack([y_outlier, x_outlier]).T

    candidate_points = np.vstack([np.hstack([yp,ym]), np.hstack([x,x])]).T

    candidate_points = np.vstack([outlier_points, candidate_points])

    yt, xt = candidate_points.T
    print xt

    #center, angle, (axis1,axis2) = fit_ellipse(candidate_points)

    fe=FitEllipse(40,100,0.0001,40)
    result = fe.ransac_fit(candidate_points)
    if result!=None:
        center, angle, (axis1,axis2) = fe.ransac_fit(candidate_points)

    print "center = ", center
    print "angle = ", angle
    print "axis1 = ", axis1
    print "axis2 = ", axis2

    fig,ax=plt.subplots(1)
    ax.plot(x,yp,'bo')
    ax.plot(x,ym,'bo')
    ax.plot(x_outlier, y_outlier, 'rx')

    from matplotlib.patches import Ellipse
    el = Ellipse(center,width=2.0*axis1,height=2.0*axis2,angle=angle,fill=False,linewidth=3,color='r')

    ax.add_artist(el)


    plt.show()

if __name__=='__main__':

    test_fit()



import logging

from allensdk.core.simple_tree import SimpleTree

from allensdk.internal.mouse_connectivity.interval_unionize.tissuecyte_unionizer import TissuecyteUnionizer
import data_utilities as du


def get_ancestor_id_map(structures):


    tree = SimpleTree( structures, 
                       lambda st: int(st['id']), 
                       lambda st: st['parent_structure_id'])
    ancestor_id_map = tree.value_map( lambda st: st['id'], 
                                      lambda st: tree.ancestor_ids([st['id']])[0] )
    for k in ancestor_id_map.keys():
        ancestor_id_map[-k] = map(lambda x: -x, ancestor_id_map[k])
      
    return ancestor_id_map


def get_volume_scale(image_resolution, voxel_depth):
    return image_resolution ** 2 * 10 ** -9 * voxel_depth


def run(input_data):

    logging.info('making ancestor id map')
    ancestor_id_map = get_ancestor_id_map(input_data['structures'])

    logging.info('computing volume scale factor')
    volume_scale = get_volume_scale(input_data['image_resolution'], input_data['reference_spacing'])  
    logging.info('volume scale factor : {0}'.format(volume_scale))

    logging.info('reference shape : {0}'.format(input_data['reference_shape']))
    logging.info('reference spacing : {0}'.format(input_data['reference_spacing']))
    logging.info('image_series_id : {0}'.format(input_data['image_series_id']))

    annotation = du.load_annotation(input_data['annotation_path'], input_data['grid_paths']['data_mask'])

    unionizer = TissuecyteUnionizer()
    unionizer.setup_interval_map(annotation)
    del annotation

    signal_arrays = du.get_injection_data(input_data['grid_paths']['injection_fraction'],
                                          input_data['grid_paths']['injection_density'], 
                                          input_data['grid_paths']['injection_energy'])
    signal_arrays.update(du.get_projection_data(input_data['grid_paths']['projection_density'],
                                                input_data['grid_paths']['projection_energy'], 
                                                input_data['grid_paths']['aav_exclusion_fraction']))
    signal_arrays.update(du.get_sum_pixels(input_data['grid_paths']['sum_pixels']))
    signal_arrays.update(du.get_sum_pixel_intensities(input_data['grid_paths']['sum_pixel_intensities'], 
                                                      input_data['grid_paths']['injection_sum_pixel_intensities']))

    for k, v in signal_arrays.iteritems():
        logging.info('sorting {0} array'.format(k))
        signal_arrays[k] = v.flat[unionizer.sort]
    
    logging.info('computing unionizes from directly annotated voxels')
    raw_unionizes = unionizer.direct_unionize(signal_arrays, pre_sorted=True)
    
    logging.info('propagating data to ancestor structures')
    raw_unionizes = TissuecyteUnionizer.propagate_unionizes(raw_unionizes, 
                                                            ancestor_id_map)

    logging.info('propagating data to bilateral unionizes')
    bilateral = TissuecyteUnionizer.propagate_to_bilateral(raw_unionizes)

    cooked_unionizes = unionizer.postprocess_unionizes(raw_unionizes, 
                                                       image_series_id=input_data['image_series_id'], 
                                                       output_spacing_iso=input_data['reference_spacing'], 
                                                       volume_scale=volume_scale, 
                                                       target_shape=input_data['reference_shape'],
                                                       sort=unionizer.sort)

    cooked_bilateral = unionizer.postprocess_unionizes(bilateral, 
                                                       image_series_id=input_data['image_series_id'], 
                                                       output_spacing_iso=input_data['reference_spacing'], 
                                                       volume_scale=volume_scale, 
                                                       target_shape=input_data['reference_shape'], 
                                                       sort=unionizer.sort)
    for item in cooked_bilateral:
        item['hemisphere_id'] = 3
        cooked_unionizes.append(item)
  
    logging.info('computed {0} unionize records'.format(len(cooked_unionizes)))
    return cooked_unionizes



import functools
import copy as cp
import logging

import numpy as np


class ImageSheet(object):


    def append(self, new_cell):
        
        if not hasattr(self, 'images'):
            self.images = [new_cell]
        else:
            self.images.append(new_cell)


    def apply(self, fn, *args, **kwargs):
        fn = functools.partial(fn, *args, **kwargs)
        self.images = map(fn, self.images)


    def copy(self):
        new_sheet = ImageSheet()
        new_sheet.images = cp.deepcopy(self.images)
        return new_sheet


    def get_output(self, axis):
        output = np.concatenate(self.images, axis=axis)
        logging.info('concatenated sheet has size: {0}'.format(output.shape))
        return output


    @staticmethod
    def build_from_image(image, n, axis):

        images = np.split(image, n, axis)

        sheet = ImageSheet()
        sheet.images = images

        return sheet



import matplotlib

matplotlib.use('agg')

import logging

import allensdk.internal.core.lims_utilities as lims_utilities
import allensdk.core.json_utilities as json_utilities

from allensdk.core.nwb_data_set import NwbDataSet
import allensdk.ephys.ephys_features as ft
from allensdk.ephys.extract_cell_features import get_square_stim_characteristics, get_ramp_stim_characteristics, get_stim_characteristics

import sys
import argparse
import os
import json
import h5py
import numpy as np

from scipy.optimize import curve_fit
import scipy.signal as sg
import scipy.misc

import datetime
import matplotlib.pyplot as plt
#import seaborn as sns

AXIS_Y_RANGE = [ -110, 60 ]

def get_time_string():
    return datetime.datetime.now().strftime("%I:%M%p %B %d, %Y")

def get_spikes(sweep_features, sweep_number):
    return get_features(sweep_features, sweep_number)["spikes"]

def get_features(sweep_features, sweep_number):
    try: 
        return sweep_features[int(sweep_number)]
    except KeyError:
        return sweep_features[str(sweep_number)]

def load_experiment(file_name, sweep_number):
    ds = NwbDataSet(file_name)
    sweep = ds.get_sweep(sweep_number)
    
    r = sweep['index_range']
    v = sweep['response'] * 1e3
    i = sweep['stimulus'] * 1e12
    dt = 1.0 / sweep['sampling_rate']
    t = np.arange(0, len(v)) * dt

    return (v, i, t, r, dt)

def plot_single_ap_values(nwb_file, sweep_numbers, rheo_features, sweep_features, cell_features, type_name):
    figs = [ plt.figure() for f in range(3+len(sweep_numbers)) ]

    v, i, t, r, dt = load_experiment(nwb_file, sweep_numbers[0])
    if type_name == "short_square" or type_name == "long_square":
        stim_start, stim_dur, stim_amp, start_idx, end_idx = get_square_stim_characteristics(i, t)
    elif type_name == "ramp":
        stim_start, start_idx = get_ramp_stim_characteristics(i, t)

    gen_features = ["threshold", "peak", "trough", "fast_trough", "slow_trough"]
    voltage_features = ["threshold_v", "peak_v", "trough_v", "fast_trough_v", "slow_trough_v"]
    time_features = ["threshold_t", "peak_t", "trough_t", "fast_trough_t", "slow_trough_t"]

    for sn in sweep_numbers:
        spikes = get_spikes(sweep_features, sn)

        if (len(spikes) < 1):
            logging.warning("no spikes in sweep %d" % sn)
            continue

        if type_name != "long_square":
            voltages = [spikes[0][f] for f in voltage_features]
            times = [spikes[0][f] for f in time_features]
        else:
            rheo_sn = cell_features["long_squares"]["rheobase_sweep"]["id"]
            rheo_spike = get_spikes(sweep_features, rheo_sn)[0]
            voltages = [ rheo_spike[f] for f in voltage_features]
            times = [ rheo_spike[f] for f in time_features]

        plt.figure(figs[0].number)
        plt.scatter(range(len(voltages)), voltages, color='gray')
        plt.tight_layout()


        plt.figure(figs[1].number)
        plt.scatter(range(len(times)), times, color='gray')
        plt.tight_layout()

        plt.figure(figs[2].number)
        plt.scatter([0], [spikes[0]['upstroke'] / (-spikes[0]['downstroke'])], color='gray')
        plt.tight_layout()


    plt.figure(figs[0].number)
    
    yvals = [float(rheo_features[k + "_v_" + type_name]) for k in gen_features if rheo_features[k + "_v_" + type_name] is not None]
    xvals = range(len(yvals))
    
    plt.scatter(xvals, yvals, color='blue', marker='_', s=40, zorder=100)
    plt.xticks(xvals, ['thr', 'pk', 'tr', 'ftr', 'str'])
    plt.title(type_name + ": voltages")

    plt.figure(figs[1].number)
    yvals = [float(rheo_features[k + "_t_" + type_name]) for k in gen_features if rheo_features[k + "_t_" + type_name] is not None]
    xvals = range(len(yvals))
    plt.scatter(xvals, yvals, color='blue', marker='_', s=40, zorder=100)
    plt.xticks(xvals, ['thr', 'pk', 'tr', 'ftr', 'str'])
    plt.title(type_name + ": times")
    
    plt.figure(figs[2].number)
    if rheo_features["upstroke_downstroke_ratio_" + type_name] is not None:
        plt.scatter([0], [float(rheo_features["upstroke_downstroke_ratio_" + type_name])], color='blue', marker='_', s=40, zorder=100)
    plt.xticks([])
    plt.title(type_name + ": up/down")

    for index, sn in enumerate(sweep_numbers):
        plt.figure(figs[3 + index].number)

        v, i, t, r, dt = load_experiment(nwb_file, sn)
        plt.plot(t, v, color='black')
        plt.title(str(sn))

        spikes = get_spikes(sweep_features, sn)

        nspikes = len(spikes)

        if type_name != "long_square" and nspikes:
            if nspikes == 0:
                logging.warning("no spikes in sweep %d" % sn)
                continue

            voltages = [spikes[0][f] for f in voltage_features]
            times = [spikes[0][f] for f in time_features]
        else:
            rheo_sn = cell_features["long_squares"]["rheobase_sweep"]["id"]
            rheo_spike = get_spikes(sweep_features, rheo_sn)[0]
            voltages = [ rheo_spike[f] for f in voltage_features ]
            times = [ rheo_spike[f] for f in time_features ]

        plt.scatter(times, voltages, color='red', zorder=20)
        

        delta_v = 5.0
        if nspikes:
            plt.plot([spikes[0]['upstroke_t'] - 1e-3 * (delta_v / spikes[0]['upstroke']),
                      spikes[0]['upstroke_t'] + 1e-3 * (delta_v / spikes[0]['upstroke'])], 
                     [spikes[0]['upstroke_v'] - delta_v, spikes[0]['upstroke_v'] + delta_v], color='red')

            plt.plot([spikes[0]['downstroke_t'] - 1e-3 * (delta_v / spikes[0]['downstroke']),
                      spikes[0]['downstroke_t'] + 1e-3 * (delta_v / spikes[0]['downstroke'])], 
                     [spikes[0]['downstroke_v'] - delta_v, spikes[0]['downstroke_v'] + delta_v], color='red')

        if type_name == "ramp":
            if nspikes:
                plt.xlim(spikes[0]["threshold_t"] - 0.002, spikes[0]["fast_trough_t"] + 0.01)
        elif type_name == "short_square":
            plt.xlim(stim_start - 0.002, stim_start + stim_dur + 0.01)
        elif type_name == "long_square":
            plt.xlim(times[0]- 0.002, times[-2] + 0.002)

        plt.tight_layout()


    return figs

#def plot_sweep_figures(nwb_file, ephys_roi_result, image_dir, sizes):
def plot_sweep_figures(nwb_file, sweep_data, image_dir, sizes):
#    try:
#        sweeps = ephys_roi_result["specimens"][0]["ephys_sweeps"]
#    except:
#        sweeps = ephys_roi_result["specimens"]["ephys_sweeps"]
    vclamp_sweep_numbers = sorted([ s['sweep_number'] for s in sweep_data if s['stimulus_units'] == 'Amps' ])

    image_file_sets = {}

    tp_set = []
    exp_set = []

    prev_sweep_number = None

    tp_len = 0.035
    tp_steps = int(tp_len * 200000)

    b, a = sg.bessel(4, 0.1, "low")

    for i, sweep_number in enumerate(vclamp_sweep_numbers):
        logging.info("plotting sweep %d" %  sweep_number)
        if i == 0:
            v_init, i_init, t_init, r_init, dt_init = load_experiment(nwb_file, sweep_number)    

            tp_fig = plt.figure()
            axTP = plt.gca()
            axTP.set_yticklabels([])
            axTP.set_xticklabels([])
            axTP.set_xlabel(str(sweep_number))
            axTP.set_ylabel('')
            xTP = t_init[0:tp_steps]
            yTP = v_init[0:tp_steps]
            axTP.plot(xTP, yTP, linewidth=1)
            axTP.set_xlim(0, tp_len)
#            sns.despine()
    
            exp_fig = plt.figure()
            axDP = plt.gca()
            axDP.set_yticklabels([])
            axDP.set_xticklabels([])
            axDP.set_xlabel(str(sweep_number))
            axDP.set_ylabel('')
            v_exp = v_init[r_init[0]:]
            t_exp = t_init[r_init[0]:]
            yDP = sg.filtfilt(b, a, v_exp, axis=0)
            xDP = t_exp
            baseline = yDP[5000:9000]
            baselineMean = np.mean(baseline)
            baselineV = (np.ones(len(xDP))) * baselineMean     
            axDP.plot(xDP, yDP, linewidth=1)
            axDP.plot(xDP, baselineV, linewidth=1)
            axDP.set_xlim(t_exp[0], t_exp[-1])
#            sns.despine()

            v_prev, i_prev, t_prev, r_prev = v_init, i_init, t_init, r_init

        else:
            v, i, t, r, dt = load_experiment(nwb_file, sweep_number)    

            tp_fig = plt.figure()
            axTP = plt.gca()
            axTP.set_yticklabels([])
            axTP.set_xticklabels([])
            axTP.set_xlabel(str(sweep_number))
            axTP.set_ylabel('')
            yTP = v[:tp_steps]
            xTP = t[:tp_steps]
            TPBL = np.mean(yTP[0:100])
            yTPN = yTP - TPBL
            yTPp = v_prev[:tp_steps]
            TPpBL = np.mean(yTPp[0:100])
            yTPpN = yTPp - TPpBL
            yTPi = v_init[:tp_steps]
            TPiBL = np.mean(yTPi[0:100])
            yTPiN = yTPi - TPiBL
            axTP.plot(xTP, yTPiN, linewidth=1)
            axTP.plot(xTP, yTPpN, linewidth=1)
            axTP.plot(xTP, yTPN, linewidth=1)
            axTP.set_xlim(0, tp_len)
#            sns.despine()

            exp_fig = plt.figure()
            axDP = plt.gca()
            axDP.set_yticklabels([])
            axDP.set_xticklabels([])
            axDP.set_xlabel(str(sweep_number))
            axDP.set_ylabel('')
            v_exp = v[r[0]:]
            t_exp = t[r[0]:]
            yDP = sg.filtfilt(b, a, v_exp, axis=0)
            xDP = t_exp
            baseline = yDP[5000:9000]
            baselineMean = np.mean(baseline)
            baselineV = (np.ones(len(xDP))) * baselineMean     
            axDP.plot(xDP, yDP, linewidth=1)
            axDP.plot(xDP, baselineV, linewidth=1)
            axDP.set_xlim(t_exp[0], t_exp[-1])
#            sns.despine()

            v_prev, i_prev, t_prev, r_prev = v, i, t, r
            
        prev_sweep_number = sweep_number

        save_figure(tp_fig, 'test_pulse_%d' % sweep_number, 'test_pulses', image_dir, sizes, image_file_sets)
        save_figure(exp_fig, 'experiment_%d' % sweep_number, 'experiments', image_dir, sizes, image_file_sets)

    return image_file_sets

def save_figure(fig, image_name, image_set_name, image_dir, sizes, image_sets, scalew=1, scaleh=1, ext='jpg'):
    plt.figure(fig.number)

    if image_set_name not in image_sets:
        image_sets[image_set_name] = { size_name: [] for size_name in sizes }

    for size_name, size in sizes.iteritems():
        fig.set_size_inches(size*scalew, size*scaleh)

        image_file = os.path.join(image_dir, "%s_%s.%s" % (image_name, size_name, ext))

        plt.savefig(image_file, bbox_inches="tight")

        image_sets[image_set_name][size_name].append(image_file)

    plt.close()


def plot_images(well_known_files, image_dir, sizes, image_sets):
    wkfs = [ f for f in well_known_files if f['filename'].endswith('tif') ]
    
    paths = [ os.path.join(f['storage_directory'], f['filename']) for f in wkfs ]

    paths = [ lims_utilities.safe_system_path(p) for p in paths ]

    image_set_name = "images"
    image_sets[image_set_name] = { size_name: [] for size_name in sizes }

    for i, path in enumerate(paths):
        image_data = plt.imread(path)
        image_data = np.array(image_data, dtype=np.float32)
        
        vmin = image_data.min()
        vmax = image_data.max()
        
        image_data = np.array((image_data - vmin) / (vmax - vmin) * 255.0, dtype=np.uint8)

        for size_name, size in sizes.iteritems():
            if size:
                s = image_data.shape
                skip = int(s[0] / size)
                sdata = image_data[::skip, ::skip]
            else:
                sdata = image_data


            filename = os.path.join(image_dir, "image_%d_%s.jpg" % (i, size_name))
            scipy.misc.imsave(filename, sdata)

            image_sets['images'][size_name].append(filename)
        

def plot_subthreshold_long_square_figures(nwb_file, cell_features, rheo_features, sweep_features, image_dir, sizes, cell_image_files):
    lsq_sweeps = cell_features["long_squares"]["sweeps"]
    sub_sweeps = cell_features["long_squares"]["subthreshold_sweeps"]
    tau_sweeps = cell_features["long_squares"]["subthreshold_membrane_property_sweeps"]

    # 0a - Plot VI curve and linear fit, along with vrest
    x = np.array([ s['stim_amp'] for s in sub_sweeps ])
    y = np.array([ s['peak_deflect'][0] for s in sub_sweeps ])
    i = np.array([ s['stim_amp'] for s in tau_sweeps ])

    fig = plt.figure()
    plt.scatter(x, y, color='black')
    plt.plot([x.min(), x.max()], [rheo_features["vrest"], rheo_features["vrest"]], color="blue", linewidth=2)
    plt.plot(i, i * 1e-3 * rheo_features["ri"] + rheo_features["vrest"], color="red", linewidth=2)
    plt.xlabel("pA")
    plt.ylabel("mV")
    plt.title("ri = {:.1f}, vrest = {:.1f}".format(rheo_features["ri"], rheo_features["vrest"]))
    plt.tight_layout()

    save_figure(fig, 'VI_curve', 'subthreshold_long_squares', image_dir, sizes, cell_image_files)
    
    # 0b - Plot tau curve and average
    fig = plt.figure()
    x = np.array([ s['stim_amp'] for s in tau_sweeps ])
    y = np.array([ s['tau'] for s in tau_sweeps ])
    plt.scatter(x, y, color='black')
    i = np.array([ s['stim_amp'] for s in tau_sweeps ])
    plt.plot([i.min(), i.max()], [cell_features["long_squares"]["tau"], cell_features["long_squares"]["tau"]], color="red", linewidth=2)
    plt.xlabel("pA")
    ylim = plt.ylim()
    plt.ylim(0, ylim[1])
    plt.ylabel("tau (s)")
    plt.tight_layout()


    save_figure(fig, 'tau_curve', 'subthreshold_long_squares', image_dir, sizes, cell_image_files)
    
    subthresh_dict = {s['id']:s for s in tau_sweeps}

    # 0c - Plot the subthreshold squares
    tau_sweeps = [ s['id'] for s in tau_sweeps ]
    tau_figs = [ plt.figure() for i in range(len(tau_sweeps)) ]

    for index, s in enumerate(tau_sweeps):
        v, i, t, r, dt = load_experiment(nwb_file, s)
        
        plt.figure(tau_figs[index].number)
        
        plt.plot(t, v, color="black")

        if index == 0:
            min_y, max_y = plt.ylim()
        else:
            ylims = plt.ylim()
            if min_y > ylims[0]:
                min_y = ylims[0]
            if max_y < ylims[1]:
                max_y = ylims[1]

        stim_start, stim_dur, stim_amp, start_idx, end_idx = get_square_stim_characteristics(i, t)
        plt.xlim(stim_start - 0.05, stim_start + stim_dur + 0.05)
        peak_idx = subthresh_dict[s]['peak_deflect'][1]
        peak_t = peak_idx*dt
        plt.scatter([peak_t], [subthresh_dict[s]['peak_deflect'][0]], color='red', zorder=10)
        popt = ft.fit_membrane_time_constant(v, t, stim_start, peak_t)
        plt.title(str(s))
        plt.plot(t[start_idx:peak_idx], exp_curve(t[start_idx:peak_idx] - t[start_idx], *popt), color='blue')


    for index, s in enumerate(tau_sweeps):
        plt.figure(tau_figs[index].number)
        plt.ylim(min_y, max_y)
        plt.tight_layout()

    for index, tau_fig in enumerate(tau_figs):
        save_figure(tau_figs[index], 'tau_%d' % index, 'subthreshold_long_squares', image_dir, sizes, cell_image_files)

def plot_short_square_figures(nwb_file, cell_features, rheo_features, sweep_features, image_dir, sizes, cell_image_files):
    repeat_amp = cell_features["short_squares"].get("stimulus_amplitude", None)

    if repeat_amp is not None:
        short_square_sweep_nums = [ s['id'] for s in cell_features["short_squares"]["common_amp_sweeps"] ]

        figs = plot_single_ap_values(nwb_file, short_square_sweep_nums, 
                                     rheo_features, sweep_features, cell_features, 
                                     "short_square") 

        for index, fig in enumerate(figs):
            save_figure(fig, 'short_squares_%d' % index, 'short_squares', image_dir, sizes, cell_image_files)

        fig = plot_instantaneous_threshold_thumbnail(nwb_file, short_square_sweep_nums, 
                                                     cell_features, rheo_features, sweep_features)

        save_figure(fig, 'instantaneous_threshold_thumbnail', 'short_squares', image_dir, sizes, cell_image_files)
                                                     

    else:
        logging.warning("No short square figures to plot.")


def plot_instantaneous_threshold_thumbnail(nwb_file, sweep_numbers, cell_features, rheo_features, sweep_features, color='red'):
    min_sweep_number = None
    for sn in sorted(sweep_numbers):
        spikes = get_spikes(sweep_features, sn)

        if len(spikes) > 0:
            min_sweep_number = sn if min_sweep_number is None else min(min_sweep_number, sn)

    fig = plt.figure(frameon=False)
    ax = plt.Axes(fig, [0., 0., 1., 1.])
    ax.set_axis_off()
    fig.add_axes(ax)    
    ax.set_yticklabels([])
    ax.set_xticklabels([])
    ax.set_xlabel('')
    ax.set_ylabel('')

    v, i, t, r, dt = load_experiment(nwb_file, sn)
    stim_start, stim_dur, stim_amp, start_idx, end_idx = get_square_stim_characteristics(i, t)

    tstart = stim_start - 0.002
    tend = stim_start + stim_dur + 0.005
    tscale = 0.005

    plt.plot(t, v, linewidth=1, color=color)
    
    plt.ylim(AXIS_Y_RANGE[0], AXIS_Y_RANGE[1])
    plt.xlim(tstart, tend)

    return fig


def plot_ramp_figures(nwb_file, sweep_info, cell_features, rheo_features, sweep_features, image_dir, sizes, cell_image_files):
    sweeps = sweep_info
    ramps_sweeps = [ s["sweep_number"] for s in sweeps if s["workflow_state"].endswith("passed") and s["ephys_stimulus"]["description"][:10] == "C1RP25PR1S"]

    figs = []
    if len(ramps_sweeps) > 0:
        figs = plot_single_ap_values(nwb_file, ramps_sweeps, rheo_features, sweep_features, cell_features, "ramp")

        for index, fig in enumerate(figs):
            save_figure(fig, 'ramps_%d' % index, 'ramps', image_dir, sizes, cell_image_files)            

def plot_rheo_figures(nwb_file, cell_features, rheo_features, sweep_features, image_dir, sizes, cell_image_files):
    rheo_sweeps = [ rheo_features["rheobase_sweep_num"] ]
    figs = plot_single_ap_values(nwb_file, rheo_sweeps, rheo_features, sweep_features, cell_features, "long_square")

    for index, fig in enumerate(figs):
        save_figure(fig, 'rheo_%d' % index, 'rheo', image_dir, sizes, cell_image_files)            

def plot_hero_figures(nwb_file, cell_features, rheo_features, sweep_features, image_dir, sizes, cell_image_files):
    fig = plt.figure()
    v, i, t, r, dt = load_experiment(nwb_file, int(rheo_features["thumbnail_sweep_num"]))
    plt.plot(t, v, color='black')
    stim_start, stim_dur, stim_amp, start_idx, end_idx = get_square_stim_characteristics(i, t)
    plt.xlim(stim_start - 0.05, stim_start + stim_dur + 0.05)
    plt.ylim(-110, 50)
    spike_times = [spk['threshold_t'] for spk in get_spikes(sweep_features, rheo_features["thumbnail_sweep_num"])]
    isis = np.diff(np.array(spike_times))
    plt.title("thumbnail {:d}, amp = {:.1f}".format(rheo_features["thumbnail_sweep_num"], stim_amp))
    plt.tight_layout()
    
    save_figure(fig, 'thumbnail_0', 'thumbnail', image_dir, sizes, cell_image_files, scalew=2)

    fig = plt.figure()
    plt.plot(range(len(isis)), isis)
    plt.ylabel("ISI (ms)")
    if rheo_features.get("adaptation", None) is not None:
        plt.title("adapt = {:.3g}".format(rheo_features["adaptation"]))
    else:
        plt.title("adapt = not defined")
    
    for k in ["has_delay", "has_burst", "has_pause"]:
        if rheo_features.get(k, None) is None:
            rheo_features[k] = False

    plt.tight_layout()
    save_figure(fig, 'thumbnail_1', 'thumbnail', image_dir, sizes, cell_image_files)
        
    yvals = [
        float(rheo_features["has_delay"]),
        float(rheo_features["has_burst"]),
        float(rheo_features["has_pause"]),
    ]
    xvals = range(len(yvals))

    fig = plt.figure()
    plt.scatter(xvals, yvals, color='red')
    plt.xticks(xvals, ['Delay', 'Burst', 'Pause'])
    plt.title("flags")
    plt.tight_layout()

    save_figure(fig, 'thumbnail_2', 'thumbnail', image_dir, sizes, cell_image_files)

    summary_fig = plot_long_square_summary(nwb_file, cell_features, rheo_features, sweep_features)
    save_figure(summary_fig, 'ephys_summary', 'thumbnail', image_dir, sizes, cell_image_files, scalew=2)


def plot_long_square_summary(nwb_file, cell_features, rheo_features, sweep_features):
    long_square_sweeps = cell_features['long_squares']['sweeps']
    long_square_sweep_numbers = [ int(s['id']) for s in long_square_sweeps ]
    
    thumbnail_summary_fig = plot_sweep_set_summary(nwb_file, int(rheo_features['thumbnail_sweep_num']), long_square_sweep_numbers)
    plt.figure(thumbnail_summary_fig.number)

    return thumbnail_summary_fig


def plot_fi_curve_figures(nwb_file, cell_features, rheo_features, sweep_features, image_dir, sizes, cell_image_files):
    fig = plt.figure()
    fi_sorted = sorted(cell_features["long_squares"]["spiking_sweeps"], key=lambda s:s['stim_amp'])
    x = [d['stim_amp'] for d in fi_sorted]
    y = [d['avg_rate'] for d in fi_sorted]
    last_zero_idx = np.nonzero(y)[0][0] - 1    
    plt.scatter(x, y, color='black')
    plt.plot(x[last_zero_idx:], cell_features["long_squares"]["fi_fit_slope"] * (np.array(x[last_zero_idx:]) - x[last_zero_idx]), color='red')
    plt.xlabel("pA")
    plt.ylabel("spikes/sec")
    plt.title("slope = {:.3g}".format(rheo_features["f_i_curve_slope"]))
    rheo_hero_sweeps = [int(rheo_features["rheobase_sweep_num"]), int(rheo_features["thumbnail_sweep_num"])]
    rheo_hero_x = []
    for s in rheo_hero_sweeps:
        v, i, t, r, dt = load_experiment(nwb_file, s)
        stim_start, stim_dur, stim_amp, start_idx, end_idx = get_square_stim_characteristics(i, t)
        rheo_hero_x.append(stim_amp)
    rheo_hero_y = [ len(get_spikes(sweep_features, s)) for s in rheo_hero_sweeps ]
    plt.scatter(rheo_hero_x, rheo_hero_y, zorder=20)
    plt.tight_layout()

    save_figure(fig, 'fi_curve', 'fi_curve', image_dir, sizes, cell_image_files, scalew=2)

def plot_sag_figures(nwb_file, cell_features, rheo_features, sweep_features, image_dir, sizes, cell_image_files):
    fig = plt.figure()
    for d in cell_features["long_squares"]["subthreshold_sweeps"]:
        if d['peak_deflect'][0] == rheo_features["vm_for_sag"]:
            v, i, t, r, dt = load_experiment(nwb_file, int(d['id']))
            stim_start, stim_dur, stim_amp, start_idx, end_idx = get_square_stim_characteristics(i, t)
            plt.plot(t, v, color='black')
            plt.scatter(d['peak_deflect'][1], d['peak_deflect'][0], color='red', zorder=10)
            #plt.plot([stim_start + stim_dur - 0.1, stim_start + stim_dur], [d['steady'], d['steady']], color='red', zorder=10)
    plt.xlim(stim_start - 0.25, stim_start + stim_dur + 0.25)
    plt.title("sag = {:.3g}".format(rheo_features['sag']))
    plt.tight_layout()

    save_figure(fig, 'sag', 'sag', image_dir, sizes, cell_image_files, scalew=2)

def mask_nulls(data):
    data[0, np.equal(data[0,:], None) | np.equal(data[0,:],0)] = np.nan

def plot_sweep_value_figures(sweep_info, image_dir, sizes, cell_image_files):
    sweeps = sorted(sweep_info, key=lambda s: s['sweep_number'] )
    
    # plot bridge balance
    data = np.array([ [ s['bridge_balance_mohm'], s['sweep_number'] ] for s in sweeps ]).T
    mask_nulls(data)

    fig = plt.figure()
    plt.title('bridge balance')
    plt.plot(data[1,:], data[0,:], marker='.')
    
    save_figure(fig, 'bridge_balance', 'sweep_values', image_dir, sizes, cell_image_files, scalew=2)

    # plot pre_vm_mv, no blowout sweep
    data = np.array([ [ s['pre_vm_mv'], s['sweep_number'] ] 
                      for s in sweeps 
                      if not s['ephys_stimulus']['description'].startswith('EXTPBLWOUT')]).T
    mask_nulls(data)

    fig = plt.figure()
    plt.title('pre vm')
    plt.plot(data[1,:], data[0,:], marker='.')
    
    save_figure(fig, 'pre_vm_mv', 'sweep_values', image_dir, sizes, cell_image_files, scalew=2)

    # plot bias current
    data = np.array([ [ s['leak_pa'], s['sweep_number'] ] for s in sweeps ]).T
    mask_nulls(data)
                   
    fig = plt.figure()
    plt.title('leak')
    plt.plot(data[1,:], data[0,:], marker='.')
    
    save_figure(fig, 'leak', 'sweep_values', image_dir, sizes, cell_image_files, scalew=2)

#def plot_cell_figures(nwb_file, ephys_roi_result, image_dir, sizes):
def plot_cell_figures(nwb_file, 
                cell_features,
                sweep_features, 
                rheo_features, 
                image_dir, 
                sweep_info,
                sizes):
    
    cell_image_files = {}

    plt.style.use('ggplot')

    logging.info("saving sweep feature figures")
    plot_sweep_value_figures(sweep_info, image_dir, sizes, cell_image_files)

    logging.info("saving tau and vi figs")
    plot_subthreshold_long_square_figures(nwb_file, cell_features, rheo_features, sweep_features, image_dir, sizes, cell_image_files)
    
    logging.info("saving short square figs")
    plot_short_square_figures(nwb_file, cell_features, rheo_features, sweep_features, image_dir, sizes, cell_image_files)

    logging.info("saving ramps")
    plot_ramp_figures(nwb_file, sweep_info, cell_features, rheo_features, sweep_features, image_dir, sizes, cell_image_files)

    logging.info("saving rheo figs")
    plot_rheo_figures(nwb_file, cell_features, rheo_features, sweep_features, image_dir, sizes, cell_image_files)

    logging.info("saving thumbnail figs")
    plot_hero_figures(nwb_file, cell_features, rheo_features, sweep_features, image_dir, sizes, cell_image_files)

    logging.info("saving fi curve figs")
    plot_fi_curve_figures(nwb_file, cell_features, rheo_features, sweep_features, image_dir, sizes, cell_image_files)

    logging.info("saving sag figs")
    plot_sag_figures(nwb_file, cell_features, rheo_features, sweep_features, image_dir, sizes, cell_image_files)

    return cell_image_files

def plot_sweep_set_summary(nwb_file, highlight_sweep_number, sweep_numbers,
                           highlight_color='#0779BE', background_color='#dddddd'):

    fig = plt.figure(frameon=False)
    ax = plt.Axes(fig, [0., 0., 1., 1.])
    ax.set_axis_off()
    fig.add_axes(ax)    
    ax.set_yticklabels([])
    ax.set_xticklabels([])
    ax.set_xlabel('')
    ax.set_ylabel('')

    for sn in sweep_numbers:
        v, i, t, r, dt = load_experiment(nwb_file, sn)
        ax.plot(t, v, linewidth=0.5, color=background_color)

    v, i, t, r, dt = load_experiment(nwb_file, highlight_sweep_number)
    plt.plot(t, v, linewidth=1, color=highlight_color)

    stim_start, stim_dur, stim_amp, start_idx, end_idx = get_square_stim_characteristics(i, t)
    
    tstart = stim_start - 0.05
    tend = stim_start + stim_dur + 0.25

    ax.set_ylim(AXIS_Y_RANGE[0], AXIS_Y_RANGE[1])
    ax.set_xlim(tstart, tend)

    return fig

def make_sweep_html(sweep_files, file_name):
    html = "<html><body>"
    html += "<a href='index.html'>Cell QC Figures</a>"

    html += "<p>page created at: %s</p>" % get_time_string()

    html += "<div style='position:absolute;width:50%;left:0;top:40'>"
    if 'test_pulses' in sweep_files:
        for small_img, large_img in zip(sweep_files['test_pulses']['small'], 
                                        sweep_files['test_pulses']['large']):
            html += "<a href='%s' target='_blank'><img src='%s'></img></a>" % ( os.path.basename(large_img), 
                                                                                os.path.basename(small_img) ) 
    html += "</div>"

    html += "<div style='position:absolute;width:50%;right:0;top:40'>"
    if 'experiments' in sweep_files:
        for small_img, large_img in zip(sweep_files['experiments']['small'], 
                                        sweep_files['experiments']['large']):
            html += "<a href='%s' target='_blank'><img src='%s'></img></a>" % ( os.path.basename(large_img),
                                                                                os.path.basename(small_img) )
    html += "</div>"
    
    html += "</body></html>"

    with open(file_name, 'w') as f:
        f.write(html)

def make_cell_html(image_files, file_name, relative_sweep_link, specimen_info, fields):

    html = "<html><body>"

    html += "<h3>Specimen %d: %s</h3>" % ( specimen_info["id"], specimen_info["name"] )
    html += "<p>page created at: %s</p>" % get_time_string()

    if relative_sweep_link:
        html += "<p><a href='sweep.html' target='_blank'> Sweep QC Figures </a></p>"
    else:
        sweep_qc_link = '/'.join([specimen_info['storage_directory'], 'qc_figures', 'sweep.html'])
        sweep_qc_link = lims_utilities.safe_system_path(sweep_qc_link)
        html += "<p><a href='%s' target='_blank'> Sweep QC Figures </a></p>" % sweep_qc_link

    fields_to_show = [ 'electrode_0_pa', 'seal_gohm', 'initial_access_resistance_mohm', 'input_resistance_mohm' ]

    html += "<table>"
    for k,v in fields.iteritems():
        html += "<tr><td>%s</td><td>%s</td></tr>" % (k, v)
    html += "</table>"

    for image_file_set_name in image_files:
        html += "<h3>%s</h3>" % image_file_set_name

        image_set_files = image_files[image_file_set_name]

        for small_img, large_img in zip(image_set_files['small'], image_set_files['large']):
            html += "<a href='%s' target='_blank'><img src='%s'></img></a>" % ( os.path.basename(large_img), 
                                                                                os.path.basename(small_img) )
    html += ("</body></html>")

    with open(file_name, 'w') as f:
        f.write(html)

def make_sweep_page(nwb_file, working_dir, sweep_data):
    sizes = { 'small': 2.0, 'large': 6.0 }

    sweep_files = plot_sweep_figures(
                    nwb_file=nwb_file, 
                    sweep_data=sweep_data, 
                    image_dir=working_dir, 
                    sizes=sizes)

    make_sweep_html(sweep_files,
                    os.path.join(working_dir, 'sweep.html'))

#def make_cell_page(nwb_file, ephys_roi_result, working_dir, save_cell_plots=True):
def make_cell_page(nwb_file, cell_features, rheo_features, sweep_features, sweep_info, well_known_files, specimen_info, working_dir, fields_to_show, save_cell_plots=True):
    """ nwb_file: name of nwb file (string)

        cell_features:

        rheo_features: dict containing extracted features from rheobase sweep

        sweep_features:

        sweep_info:

        well_known_files: LIMS-output information containing graphics
        file names

        working_dir:

        save_cell_plots:

    """

    if save_cell_plots:
        sizes = { 'small': 2.0, 'large': 6.0 }
        cell_files = plot_cell_figures(
                        nwb_file = nwb_file,
                        cell_features = cell_features,
                        rheo_features = rheo_features,
                        sweep_features = sweep_features,
                        sweep_info = sweep_info,
                        image_dir = working_dir,
                        sizes = sizes)
    else:
        cell_files = {}
        
    logging.info("saving images")
    sizes = { 'small': 200, 'large': None }
    plot_images(well_known_files, working_dir, sizes, cell_files)

    sweep_page = os.path.join(working_dir, 'sweep.html')
    relative_sweep_link = os.path.exists(sweep_page)

    if not relative_sweep_link:
        logging.info("sweep page doesn't exist, point to production sweep page")
    
    make_cell_html(cell_files,
                   os.path.join(working_dir, 'index.html'),
                   relative_sweep_link,
                   specimen_info,
                   fields_to_show)

def exp_curve(x, a, inv_tau, y0):
    ''' Function used for tau curve fitting '''
    return y0 + a * np.exp(-inv_tau * x)


#def main():
#    parser = argparse.ArgumentParser(description='analyze specimens for cell-wide features')
#    parser.add_argument('nwb_file')
#    parser.add_argument('feature_json')
#    parser.add_argument('--output_directory', default='.')
#    parser.add_argument('--no-sweep-page', action='store_false', dest='sweep_page')
#    parser.add_argument('--no-cell-page', action='store_false', dest='cell_page')
#    parser.add_argument('--log_level')
#    
#
#    args = parser.parse_args()
#
#    if args.log_level:
#        logging.getLogger().setLevel(args.log_level)
#
#    ephys_roi_result = json_utilities.read(args.feature_json)
#
#    if args.sweep_page:
#        logging.debug("making sweep page")
#        make_sweep_page(args.nwb_file, ephys_roi_result, args.output_directory)
#
#    if args.cell_page:
#        logging.debug("making cell page")
#        make_cell_page(args.nwb_file, ephys_roi_result, args.output_directory, True)
#
#
#
#if __name__ == '__main__': main()



from __future__ import division

import numpy as np

from unionize_record import Unionize


class CavUnionize(Unionize):

    __slots__ = ['sum_pixels', 'sum_cav_pixels']

    def __init__(self, *args, **kwargs):
        for key in self.__slots__:
            setattr(self, key, 0)
        
        
    def calculate(self, low, high, data_arrays):
        data_arrays = self.slice_arrays(low, high, data_arrays)

        self.sum_pixels = data_arrays['sum_pixels'].sum()
        self.sum_cav_pixels = np.multiply(data_arrays['sum_pixels'], data_arrays['cav_density']).sum()
        
        
    def propagate(self, ancestor):
        ancestor.sum_pixels += self.sum_pixels
        ancestor.sum_cav_pixels += self.sum_cav_pixels
        return ancestor
        
        
    def output(self, volume_scale):
        return {'structure_volume': self.sum_pixels * volume_scale, 
                'signal_volume': self.sum_cav_pixels * volume_scale, 
                'signal_density': self.sum_cav_pixels / self.sum_pixels if self.sum_pixels > 0 else 0}




import os
import shutil
import subprocess
import logging
import logging.config as lc
import allensdk.core.json_utilities as ju
from allensdk.core.nwb_data_set import NwbDataSet
from pkg_resources import resource_filename  # @UnresolvedImport
from allensdk.internal.api.queries.optimize_config_reader import OptimizeConfigReader
from allensdk.model.biophys_sim.config import Config
from allensdk.internal.model.biophysical.make_deap_fit_json import Report
from allensdk.internal.api.queries.biophysical_module_api import BiophysicalModuleApi
import allensdk.model.biophysical as hoc_location


class RunOptimize(object):
    _log = logging.getLogger('allensdk.internal.model.biophysical.run_optimize')

    def __init__(self,
                 input_json,
                 output_json):
        self.input_json = input_json
        self.output_json = output_json
        self.app_config = None
        self.manifest = None
        self.data_set = None


    def load_manifest(self):
        self.app_config = Config().load(self.input_json)
        self.manifest = self.app_config.manifest
        self.data_set = NwbDataSet(self.manifest.get_path('stimulus_path'))


    def nrnivmodl(self):
        RunOptimize._log.debug("nrnivmodl")

        subprocess.call(['nrnivmodl', './modfiles'])


    def info(self, lims_json_path):
        ''' return a string that a bash script can use
        to find the working directory, etc. to clean up.
        '''
        ocr = OptimizeConfigReader()
        ocr.read_lims_file(lims_json_path)
        
        print(self.app_config.data['runs'][0]['specimen_id'])
        print(self.manifest.get_path('BASEDIR'))


    def copy_local(self):
        '''
        Note
        ----
        For files that aren't needed for local debugging, use write_manifest instead.
        '''
        self.load_manifest()

        modfile_dir = self.manifest.get_path('MODFILE_DIR')

        if not os.path.exists(modfile_dir):
            os.mkdir(modfile_dir)

        output_dir = self.manifest.get_path('WORKDIR')
        if not os.path.exists(output_dir):
            os.mkdir(output_dir)

        modfiles = [self.manifest.get_path(key) for key,info
                    in self.manifest.path_info.items()
                    if 'format' in info and info['format'] == 'MODFILE']
        for from_file in modfiles:
            RunOptimize._log.debug("copying %s to %s" % (from_file, modfile_dir))
            shutil.copy(from_file, modfile_dir)

        shutil.copy(resource_filename(hoc_location.__name__,
                                      'cell.hoc'),
                    self.manifest.get_path('BASEDIR'))


    def generate_manifest_rma(self,
                              neuronal_model_id,
                              manifest_path,
                              api_url=None):
        '''
        Note
        ----
        Other necessary files are also written.
        '''
        import json
        from allensdk.api.api import Api

        bma = BiophysicalModuleApi(api_url)
        data = bma.get_neuronal_models(neuronal_model_id)

        ocr = OptimizeConfigReader()
        ocr.read_lims_message(data, 'lims_message.json')

        with open('lims_message.json', 'w') as f:
            f.write(json.dumps(data[0], sort_keys=True, indent=2))

        ocr.to_manifest(manifest_path)


    def generate_manifest_lims(self,
                               lims_json_path,
                               manifest_path):
        '''
        Note
        ----
        Other necessary files are also written.
        '''
        ocr = OptimizeConfigReader()
        ocr.read_lims_file(lims_json_path)

        ocr.to_manifest(manifest_path)


    def start_specimen(self):
        import allensdk.internal.model.biophysical.run_passive_fit as run_passive_fit
        import allensdk.internal.model.biophysical.fit_stage_1 as fit_stage_1
        import allensdk.internal.model.biophysical.fit_stage_2 as fit_stage_2

        self.load_manifest()
        
        self.passive_fit_data = \
            run_passive_fit.run_passive_fit(self.app_config)

        ju.write(self.manifest.get_path('passive_fit_data'),
                 self.passive_fit_data)

        self.stage_1_jobs = \
            fit_stage_1.prepare_stage_1(self.app_config,
                                        self.passive_fit_data)

        ju.write(self.manifest.get_path('stage_1_jobs'),
                 self.stage_1_jobs)
        fit_stage_1.run_stage_1(self.stage_1_jobs)

        output_directory = self.manifest.get_path('WORKDIR')
        stage_2_jobs = fit_stage_2.prepare_stage_2(output_directory)
        fit_stage_2.run_stage_2(stage_2_jobs)


    def make_fit(self):
        self.load_manifest()
        
        fit_types = ["f9", "f13"]

        best_fit_values = { fit_type: None for fit_type in fit_types }

        specimen_id = self.app_config.data['runs'][0]['specimen_id']

        for fit_type in fit_types:
            fit_type_dir = self.manifest.get_path('fit_type_path', fit_type)

            if os.path.exists(fit_type_dir):
                report = Report(self.app_config,
                                fit_type)
                report.generate_fit_file()
                if fit_type in best_fit_values.keys():
                    best_fit_values[fit_type] = report.best_fit_value()

        best_fit_type, min_fit_value = reduce(lambda a, b: a if (a[1] < b[1]) else b,
                                              (i for i in best_fit_values.items() if i[1] is not None))        
        best_fit_file = self.manifest.get_path('output_fit_file',
                                               specimen_id,
                                               best_fit_type)

        lims_upload_config = OptimizeConfigReader()
        lims_upload_config.read_json(self.manifest.get_path('neuronal_model_data'))

        lims_upload_config.update_well_known_file(best_fit_file,
                                                  well_known_file_type_id=OptimizeConfigReader.NEURONAL_MODEL_PARAMETERS)
        lims_upload_config.write_file(output_json)


def main(command, input_json, output_json):
    ''' Entry point for module.
        :param command: select behavior, nrnivmodl or simulate
        :type command: string    
        :param lims_strategy_json: path to json file output from lims.
        :type lims_strategy_json: string
        :param lims_response_json: path to json file returned to lims.
        :type lims_response_json: string
    '''

    o = RunOptimize(input_json,
                    output_json)

    if 'LOG_CFG' in os.environ:
        log_config = os.environ['LOG_CFG']
    else:
        log_config = resource_filename('allensdk.model.biophysical',
                                       'logging.conf')
        os.environ['LOG_CFG'] = log_config
    lc.fileConfig(log_config)

    if 'nrnivmodl' == command:
        o.nrnivmodl()
    elif 'info' == command:
        o.info(input_json)
    elif 'generate_manifest_rma' == command:
        o.generate_manifest_rma(input_json, output_json)
    elif 'generate_manifest_lims' == command:
        o.generate_manifest_lims(input_json, output_json)
    elif 'write_manifest' == command:
        o.write_manifest()
    elif 'copy_local' == command:
        o.copy_local()
    elif 'start_specimen' == command:
        o.start_specimen()
    elif 'make_fit' == command:
        o.make_fit()
    else:
        RunOptimize._log.error("no command")

    print('done')


if __name__ == '__main__':
    import sys

    command, input_json, output_json = sys.argv[-3:]

    RunOptimize._log.debug("command: %s" % (command))
    RunOptimize._log.debug("input json: %s" % (input_json))
    RunOptimize._log.debug("output json: %s" % (output_json))

    main(command, input_json, output_json)

    RunOptimize._log.debug("success")



import argparse
import json
import logging

import numpy as np
import nrrd

from allensdk.internal.core.lims_pipeline_module import PipelineModule, run_module
from allensdk.internal.mouse_connectivity.tissuecyte_gridding.run_cav import run_cav
from allensdk.internal.mouse_connectivity.tissuecyte_gridding.cav_subimage import CavSubImage


def main():

    mod = PipelineModule()
    mod.parser.add_argument("--nprocesses", type=int, default=1)
    mod.parser.add_argument("--reduce_level", type=int, default=0)
    mod.parser.add_argument("--filter_bit", type=int, default=None)
    args = mod.parser.parse_args()

    logging.info("reading input")
    input_data = mod.input_data()

    sub_images = input_data['sub_images']

    input_dimensions = [sub_images[0]['dimensions']['column'], 
                        sub_images[0]['dimensions']['row'], 
                        input_data['sub_image_count']]  

    input_spacing = [sub_images[0]['spacing']['column'], 
                     sub_images[0]['spacing']['row'], 
                     input_data['image_series_slice_spacing']] 


    for ii, si in enumerate(sub_images):
        del si['dimensions']
        del si['spacing']
        si['polygon_info'] = si['polygons']
        del si['polygons']
    sub_images = sorted(sub_images, key=lambda si: si['specimen_tissue_index'])

    output_dimensions = [input_data['reference_dimensions']['slice'], 
                         input_data['reference_dimensions']['row'], 
                         input_data['reference_dimensions']['column']]

    output_spacing = [input_data['reference_spacing']['slice'], 
                      input_data['reference_spacing']['row'], 
                      input_data['reference_spacing']['column']]

    subimage_kwargs = {'cls': CavSubImage}
    if args.filter_bit is not None:
        subimage_kwargs['filter_bit'] = args.filter_bit

    paths = run_cav(grid_prefix=input_data['grid_prefix'], 
                    accumulator_prefix=input_data['accumulator_prefix'], 
                    in_dims=input_dimensions, 
                    in_spacing=input_spacing, 
                    out_dims=output_dimensions, 
                    out_spacing=output_spacing, 
                    reduce_level=args.reduce_level, 
                    subimages=sub_images, 
                    subimage_kwargs=subimage_kwargs, 
                    nprocesses=args.nprocesses, 
                    affine_params=input_data['affine_params'], 
                    dfmfld_path=input_data['deformation_field_path'])
            

    logging.info('writing output')
    mod.write_output_data({'output_file_paths': paths})


if __name__ == '__main__':
    
#    logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    logging.getLogger('').setLevel(logging.INFO)
    
    main()



import argparse, sys, logging

import allensdk.core.json_utilities as ju
import allensdk.internal.model.glif.find_sweeps as fs

from allensdk.internal.model.glif.glif_optimizer_neuron import GlifOptimizerNeuron
from allensdk.internal.model.glif.glif_experiment import GlifExperiment
from allensdk.internal.model.glif.glif_optimizer import GlifOptimizer

from allensdk.internal.model.data_access import load_sweeps
from allensdk.internal.model.glif.find_spikes import find_spikes_list
import allensdk.core.json_utilities as ju
import allensdk.internal.model.glif.preprocess_neuron as pn

def get_optimize_sweep_numbers(sweep_index): 
    #TODO: why is this here--why are sweep indicies being fed to a find_noise_sweeps sweeps and specifying
    #noise?--shouldn't the sweeps already be provided?   
    return fs.find_noise_sweeps(sweep_index)['noise1']

def optimize_neuron(model_config, sweep_index, nwb_file, save_callback=None):
    '''Optimizes a neuron.  
    1. Loads optimizer and neuron configuration data.
    2. Loads the voltage trace sweeps that will be optimized
    3. Configures the experiment and optimizer
    4. Runs the optimizer
    5. TODO: where is data saved
    
    Parameters
    ----------
    model_config : dictionary
        contains values of neuron and optimizer parameters
    sweep_index : list of integers
        indices (as labeled in the data configuration file) of sweeps that will be optimized
    save_callback : module
        saves output
    '''
    # define the neuron and optimizer dictionaries from the model configuration
    neuron_config = model_config['neuron']
    optimizer_config = model_config['optimizer']

    # load the neuron with along with the methods needed for optimization
    neuron = GlifOptimizerNeuron.from_dict(neuron_config)

    # TODO: not sure what this is doing
    optimize_sweeps = get_optimize_sweep_numbers(sweep_index)
    
    # load the sweeps to be optimized
    optimize_data = load_sweeps(nwb_file, optimize_sweeps, neuron.dt, 
                                optimizer_config["cut"], optimizer_config["bessel"])

    # needed to offset all voltages by El_reference
    El_reference = neuron_config['El_reference']

    # get indicies of spikes and voltage at those spikes
    spike_ind, spike_v = find_spikes_list(optimize_data['voltage'], neuron_config['dt'])

    # get times of spikes 
    grid_spike_times = [ si*neuron_config['dt'] for si in spike_ind ]
    
    # convert voltage at spikes into reference frame of El
    grid_spike_voltages_in_ref_to_zero = [ sv - El_reference for sv in spike_v ]

    # convert voltage into reference frame of El 
    resp_list = [ d - El_reference for d in optimize_data['voltage'] ]  

    # configure experiment
    experiment = GlifExperiment(neuron = neuron, 
                                dt = neuron.dt,
                                stim_list = optimize_data['current'],
                                resp_list = resp_list,
                                spike_time_steps = spike_ind, 
                                grid_spike_times = grid_spike_times,
                                grid_spike_voltages = grid_spike_voltages_in_ref_to_zero,
                                param_fit_names = optimizer_config['param_fit_names'])

    # configure optimizer
    optimizer = GlifOptimizer(experiment = experiment, 
                              dt = neuron.dt,              
                              outer_iterations = optimizer_config['outer_iterations'],
                              inner_iterations = optimizer_config['inner_iterations'],
                              sigma_inner = optimizer_config['sigma_inner'],
                              sigma_outer = optimizer_config['sigma_outer'],
                              param_fit_names = optimizer_config['param_fit_names'],
                              stim = optimize_data['current'],
                              error_function_data = optimizer_config['error_function_data'],
                              xtol = optimizer_config['xtol'],
                              ftol = optimizer_config['ftol'],
                              internal_iterations = optimizer_config['internal_iterations'],
                              init_params = optimizer_config.get('init_params', None),
                              bessel = optimizer_config['bessel'])

    def save(optimizer, outer, inner):
        logging.info('finished outer: %d inner: %d' % (outer, inner))
        if save_callback:
            save_callback(optimizer, outer, inner)
    
    # run the optimizer
    best_param, begin_param = optimizer.run_many(save) 

    # over write the the initial experiment parameters with the best found parameters
    # TODO: but why do this since it is not being returned
    experiment.set_neuron_parameters(best_param)
    
    return optimizer, best_param, begin_param
        
def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('model_config_file')
    parser.add_argument('sweeps_file')
    parser.add_argument('output_file')
    parser.add_argument("--dt", default=pn.DEFAULT_DT)
    parser.add_argument("--bessel", default=pn.DEFAULT_BESSEL)
    parser.add_argument("--cut", default=pn.DEFAULT_CUT)

    args = parser.parse_args()

    model_config = ju.read(args.model_config_file)
    sweep_list = ju.read(args.sweeps_file)

    sweep_index = { s['sweep_number']:s for s in sweep_list }

    try:
        neuron, best_param, begin_param = optimize_neuron(model_config, sweep_index, dt, cut, bessel)
        ju.write(args.output_file, neuron.to_dict())
    except Exception, e:
        logging.error(e.message)

if __name__ == "__main__": main()
